# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast

from oslo_log import log as logging

from cinder import exception
from cinder import interface
from cinder.volume import driver
from cinder.volume.drivers.dell_emc.powermax import common
from cinder.volume.drivers.san import san
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)


@interface.volumedriver
class PowerMaxFCDriver(san.SanDriver, driver.FibreChannelDriver):
    """FC Drivers for PowerMax using REST.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        1.1.0 - Multiple pools and thick/thin provisioning,
                performance enhancement.
        2.0.0 - Add driver requirement functions
        2.1.0 - Add consistency group functions
        2.1.1 - Fixed issue with mismatched config (bug #1442376)
        2.1.2 - Clean up failed clones (bug #1440154)
        2.1.3 - Fixed a problem with FAST support (bug #1435069)
        2.2.0 - Add manage/unmanage
        2.2.1 - Support for SE 8.0.3
        2.2.2 - Update Consistency Group
        2.2.3 - Pool aware scheduler(multi-pool) support
        2.2.4 - Create CG from CG snapshot
        2.3.0 - Name change for MV and SG for FAST (bug #1515181)
              - Fix for randomly choosing port group. (bug #1501919)
              - get_short_host_name needs to be called in find_device_number
                (bug #1520635)
              - Proper error handling for invalid SLOs (bug #1512795)
              - Extend Volume for VMAX3, SE8.1.0.3
              https://blueprints.launchpad.net/cinder/+spec/vmax3-extend-volume
              - Incorrect SG selected on an attach (#1515176)
              - Cleanup Zoning (bug #1501938)  NOTE: FC only
              - Last volume in SG fix
              - _remove_last_vol_and_delete_sg is not being called
                for VMAX3 (bug #1520549)
              - necessary updates for CG changes (#1534616)
              - Changing PercentSynced to CopyState (bug #1517103)
              - Getting iscsi ip from port in existing masking view
              - Replacement of EMCGetTargetEndpoints api (bug #1512791)
              - VMAX3 snapvx improvements (bug #1522821)
              - Operations and timeout issues (bug #1538214)
        2.4.0 - EMC VMAX - locking SG for concurrent threads (bug #1554634)
              - SnapVX licensing checks for VMAX3 (bug #1587017)
              - VMAX oversubscription Support (blueprint vmax-oversubscription)
              - QoS support (blueprint vmax-qos)
        2.5.0 - Attach and detach snapshot (blueprint vmax-attach-snapshot)
              - MVs and SGs not reflecting correct protocol (bug #1640222)
              - Storage assisted volume migration via retype
                (bp vmax-volume-migration)
              - Support for compression on All Flash
              - Volume replication 2.1 (bp add-vmax-replication)
              - rename and restructure driver (bp vmax-rename-dell-emc)
        3.0.0 - REST based driver
              - Retype (storage-assisted migration)
              - QoS support
              - Support for compression on All Flash
              - Support for volume replication
              - Support for live migration
              - Support for Generic Volume Group
        3.1.0 - Support for replication groups (Tiramisu)
              - Deprecate backend xml configuration
              - Support for async replication (vmax-replication-enhancements)
              - Support for SRDF/Metro (vmax-replication-enhancements)
              - Support for manage/unmanage snapshots
                (vmax-manage-unmanage-snapshot)
              - Support for revert to volume snapshot
        3.2.0 - Support for retyping replicated volumes (bp
                vmax-retype-replicated-volumes)
              - Support for multiattach volumes (bp vmax-allow-multi-attach)
              - Support for list manageable volumes and snapshots
                (bp/vmax-list-manage-existing)
              - Fix for SSL verification/cert application (bug #1772924)
              - Log VMAX metadata of a volume (bp vmax-metadata)
              - Fix for get-pools command (bug #1784856)
        4.0.0 - Fix for initiator retrieval and short hostname unmapping
                (bugs #1783855 #1783867)
              - Fix for HyperMax OS Upgrade Bug (bug #1790141)
              - Support for failover to secondary Unisphere
                (bp/vmax-unisphere-failover)
              - Rebrand from VMAX to PowerMax(bp/vmax-powermax-rebrand)
              - Change from 84 to 90 REST endpoints (bug #1808539)
              - Fix for PowerMax OS replication settings (bug #1812685)
              - Support for storage-assisted in-use retype
                (bp/powermax-storage-assisted-inuse-retype)
        4.1.0 - Changing from 90 to 91 rest endpoints
              - Support for Rapid TDEV Delete (bp powermax-tdev-deallocation)
              - PowerMax OS Metro formatted volumes fix (bug #1829876)
              - Support for Metro ODE (bp/powermax-metro-ode)
              - Removal of san_rest_port from PowerMax cinder.conf config
              - SnapVX noCopy mode enabled for all links
              - Volume/Snapshot backed metadata inclusion
              - Debug metadata compression and service level info fix
        4.2.0 - Support of Unisphere storage group and array tags
              - User defined override for short host name and port group name
                (bp powermax-user-defined-hostname-portgroup)
              - Switch to Unisphere REST API public replication endpoints
              - Support for multiple replication devices
              - Pools bug fix allowing 'None' variants (bug #1873253)
        4.3.0 - Changing from 91 to 92 REST endpoints
              - Support for Port Group and Port load balancing
                (bp powermax-port-load-balance)
              - Fix to enable legacy volumes to live migrate (#1867163)
              - Use of snap id instead of generation (bp powermax-snapset-ids)
              - Support for Failover Abilities (bp/powermax-failover-abilities)
        4.3.1 - Fix non-temporary snapshot delete (#1887962)
    """

    VERSION = "4.3.1"

    # ThirdPartySystems wiki
    CI_WIKI_NAME = "EMC_VMAX_CI"

    def __init__(self, *args, **kwargs):

        super(PowerMaxFCDriver, self).__init__(*args, **kwargs)
        self.active_backend_id = kwargs.get('active_backend_id', None)
        self.common = common.PowerMaxCommon(
            'FC',
            self.VERSION,
            configuration=self.configuration,
            active_backend_id=self.active_backend_id)
        self.performance = self.common.performance
        self.rest = self.common.rest
        self.zonemanager_lookup_service = fczm_utils.create_lookup_service()

    @classmethod
    def get_driver_options(cls):
        additional_opts = cls._get_oslo_driver_opts(
            'san_ip', 'san_login', 'san_password', 'driver_ssl_cert_verify',
            'max_over_subscription_ratio', 'reserved_percentage',
            'replication_device')
        return common.powermax_opts + additional_opts

    def check_for_setup_error(self):
        pass

    def create_volume(self, volume):
        """Creates a PowerMax/VMAX volume.

        :param volume: the cinder volume object
        :returns: provider location dict
        """
        return self.common.create_volume(volume)

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot.

        :param volume: the cinder volume object
        :param snapshot: the cinder snapshot object
        :returns: provider location dict
        """
        return self.common.create_volume_from_snapshot(
            volume, snapshot)

    def create_cloned_volume(self, volume, src_vref):
        """Creates a cloned volume.

        :param volume: the cinder volume object
        :param src_vref: the source volume reference
        :returns: provider location dict
        """
        return self.common.create_cloned_volume(volume, src_vref)

    def delete_volume(self, volume):
        """Deletes a PowerMax/VMAX volume.

        :param volume: the cinder volume object
        """
        self.common.delete_volume(volume)

    def create_snapshot(self, snapshot):
        """Creates a snapshot.

        :param snapshot: the cinder snapshot object
        :returns: provider location dict
        """
        src_volume = snapshot.volume
        return self.common.create_snapshot(snapshot, src_volume)

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot.

        :param snapshot: the cinder snapshot object
        """
        src_volume = snapshot.volume
        self.common.delete_snapshot(snapshot, src_volume)

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume.

        :param context: the context
        :param volume: the cinder volume object
        """
        pass

    def create_export(self, context, volume, connector):
        """Driver entry point to get the export info for a new volume.

        :param context: the context
        :param volume: the cinder volume object
        :param connector: the connector object
        """
        pass

    def remove_export(self, context, volume):
        """Driver entry point to remove an export for a volume.

        :param context: the context
        :param volume: the cinder volume object
        """
        pass

    @staticmethod
    def check_for_export(context, volume_id):
        """Make sure volume is exported.

        :param context: the context
        :param volume_id: the volume id
        """
        pass

    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection info.

        Assign any created volume to a compute node/host so that it can be
        used from that host.

        The  driver returns a driver_volume_type of 'fibre_channel'.
        The target_wwn can be a single entry or a list of wwns that
        correspond to the list of remote wwn(s) that will export the volume.
        Example return values:

        .. code-block:: json

            {
                'driver_volume_type': 'fibre_channel'
                'data': {
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': '1234567890123',
                }
            }

            or

            {
                'driver_volume_type': 'fibre_channel'
                'data': {
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': ['1234567890123', '0987654321321'],
                }
            }

        :param volume: the cinder volume object
        :param connector: the connector object
        :returns: dict -- the target_wwns and initiator_target_map
        """
        device_info = self.common.initialize_connection(
            volume, connector)
        if device_info:
            conn_info = self.populate_data(device_info, volume, connector)
            fczm_utils.add_fc_zone(conn_info)
            return conn_info
        else:
            return {}

    def populate_data(self, device_info, volume, connector):
        """Populate data dict.

        Add relevant data to data dict, target_lun, target_wwn and
        initiator_target_map.
        :param device_info: device_info
        :param volume: the volume object
        :param connector: the connector object
        :returns: dict -- the target_wwns and initiator_target_map
        """
        device_number = device_info['hostlunid']
        target_wwns, init_targ_map = self._build_initiator_target_map(
            volume, connector, device_info)

        data = {'driver_volume_type': 'fibre_channel',
                'data': {'target_lun': device_number,
                         'target_discovered': True,
                         'target_wwn': target_wwns,
                         'initiator_target_map': init_targ_map}}

        LOG.debug("Return FC data for zone addition: %(data)s.",
                  {'data': data})

        return data

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector.

        Return empty data if other volumes are in the same zone.
        The FibreChannel ZoneManager doesn't remove zones
        if there isn't an initiator_target_map in the
        return of terminate_connection.

        :param volume: the volume object
        :param connector: the connector object
        :returns: dict -- the target_wwns and initiator_target_map if the
            zone is to be removed, otherwise empty
        """
        data = {'driver_volume_type': 'fibre_channel', 'data': {}}
        zoning_mappings = {}
        if connector:
            zoning_mappings = self._get_zoning_mappings(volume, connector)

        if zoning_mappings:
            self.common.terminate_connection(volume, connector)
            data = self._cleanup_zones(zoning_mappings)
        fczm_utils.remove_fc_zone(data)
        return data

    def _get_zoning_mappings(self, volume, connector):
        """Get zoning mappings by building up initiator/target map.

        :param volume: the volume object
        :param connector: the connector object
        :returns: dict -- the target_wwns and initiator_target_map if the
            zone is to be removed, otherwise empty
        """
        loc = volume.provider_location
        name = ast.literal_eval(loc)
        host_label = self.common.utils.get_host_name_label(
            connector['host'], self.common.powermax_short_host_name_template)
        zoning_mappings = {}
        try:
            array = name['array']
            device_id = name['device_id']
        except KeyError:
            array = name['keybindings']['SystemName'].split('+')[1].strip('-')
            device_id = name['keybindings']['DeviceID']
        LOG.debug("Start FC detach process for volume: %(volume)s.",
                  {'volume': volume.name})

        masking_views, is_metro = (
            self.common.get_masking_views_from_volume(
                array, volume, device_id, host_label))
        if not masking_views:
            # Backward compatibility with pre Ussuri short host name.
            host_label = self.common.utils.get_host_short_name(
                connector['host'])
            masking_views, is_metro = (
                self.common.get_masking_views_from_volume(
                    array, volume, device_id, host_label))
        if masking_views:
            portgroup = (
                self.common.get_port_group_from_masking_view(
                    array, masking_views[0]))
            initiator_group = (
                self.common.get_initiator_group_from_masking_view(
                    array, masking_views[0]))

            LOG.debug("Found port group: %(portGroup)s "
                      "in masking view %(maskingView)s.",
                      {'portGroup': portgroup,
                       'maskingView': masking_views[0]})
            # Map must be populated before the terminate_connection
            target_wwns, init_targ_map = self._build_initiator_target_map(
                volume, connector)
            zoning_mappings = {'port_group': portgroup,
                               'initiator_group': initiator_group,
                               'target_wwns': target_wwns,
                               'init_targ_map': init_targ_map,
                               'array': array}
        if is_metro:
            rep_data = volume.replication_driver_data
            name = ast.literal_eval(rep_data)
            try:
                metro_array = name['array']
                metro_device_id = name['device_id']
            except KeyError:
                LOG.error("Cannot get remote Metro device information "
                          "for zone cleanup. Attempting terminate "
                          "connection...")
            else:
                masking_views, __ = (
                    self.common.get_masking_views_from_volume(
                        metro_array, volume, metro_device_id, host_label))
                if masking_views:
                    metro_portgroup = (
                        self.common.get_port_group_from_masking_view(
                            metro_array, masking_views[0]))
                    metro_ig = (
                        self.common.get_initiator_group_from_masking_view(
                            metro_array, masking_views[0]))
                    zoning_mappings.update(
                        {'metro_port_group': metro_portgroup,
                         'metro_ig': metro_ig, 'metro_array': metro_array})
        if not masking_views:
            LOG.warning("Volume %(volume)s is not in any masking view.",
                        {'volume': volume.name})
        return zoning_mappings

    def _cleanup_zones(self, zoning_mappings):
        """Cleanup zones after terminate connection.

        :param zoning_mappings: zoning mapping dict
        :returns: data - dict
        """
        data = {'driver_volume_type': 'fibre_channel', 'data': {}}
        try:
            LOG.debug("Looking for masking views still associated with "
                      "Port Group %s.", zoning_mappings['port_group'])
            masking_views = self.common.get_common_masking_views(
                zoning_mappings['array'], zoning_mappings['port_group'],
                zoning_mappings['initiator_group'])
        except (KeyError, ValueError, TypeError):
            masking_views = []

        if masking_views:
            LOG.debug("Found %(numViews)d MaskingViews.",
                      {'numViews': len(masking_views)})
        else:  # no masking views found
            # Check if there any Metro masking views
            if zoning_mappings.get('metro_array'):
                masking_views = self.common.get_common_masking_views(
                    zoning_mappings['metro_array'],
                    zoning_mappings['metro_port_group'],
                    zoning_mappings['metro_ig'])
            if not masking_views:
                LOG.debug("No MaskingViews were found. Deleting zone.")
                data = {'driver_volume_type': 'fibre_channel',
                        'data': {'target_wwn': zoning_mappings['target_wwns'],
                                 'initiator_target_map':
                                     zoning_mappings['init_targ_map']}}

                LOG.debug("Return FC data for zone removal: %(data)s.",
                          {'data': data})

        return data

    def _build_initiator_target_map(self, volume, connector, device_info=None):
        """Build the target_wwns and the initiator target map.

        :param volume: the cinder volume object
        :param connector: the connector object
        :param device_info: device_info
        :returns: target_wwns -- list, init_targ_map -- dict
        """
        target_wwns, init_targ_map = [], {}
        initiator_wwns = connector['wwpns']
        fc_targets, metro_fc_targets = (
            self.common.get_target_wwns_from_masking_view(
                volume, connector))

        # If load balance is enabled we want to select only the FC target that
        # has the lowest load of all ports in selected port group.
        # Note: device_info in if condition as this method is called also for
        # terminate connection, we only want to calculate load on initialise
        # connection.
        if device_info and self.performance.config.get('load_balance'):
            try:
                array_id = device_info.get('array')
                masking_view = device_info.get('maskingview')
                # Get PG from MV
                port_group = self.rest.get_element_from_masking_view(
                    array_id, masking_view, portgroup=True)
                # Get port list from PG
                port_list = self.rest.get_port_ids(array_id, port_group)
                # Get lowest load port in PG
                load, metric, port = self.performance.process_port_load(
                    array_id, port_list)
                LOG.info("Lowest %(met)s load port is %(port)s: %(load)s",
                         {'met': metric, 'port': port, 'load': load})
                # Get target WWN
                port_details = self.rest.get_port(array_id, port)
                port_info = port_details.get('symmetrixPort')
                port_wwn = port_info.get('wwn_node')
                LOG.info("Port %(p)s WWN: %(wwn)s",
                         {'p': port, 'wwn': port_wwn})
                # Set lowest load port WWN as FC target for connection
                fc_targets = [port_wwn]
            except exception.VolumeBackendAPIException:
                LOG.error("There was an error calculating port load, "
                          "reverting to default target selection.")
                fc_targets, __ = (
                    self.common.get_target_wwns_from_masking_view(
                        volume, connector))

        if self.zonemanager_lookup_service:
            fc_targets.extend(metro_fc_targets)
            mapping = (
                self.zonemanager_lookup_service.
                get_device_mapping_from_network(initiator_wwns, fc_targets))
            for entry in mapping:
                map_d = mapping[entry]
                target_wwns.extend(map_d['target_port_wwn_list'])
                for initiator in map_d['initiator_port_wwn_list']:
                    init_targ_map[initiator] = map_d['target_port_wwn_list']
        else:  # No lookup service, pre-zoned case.
            target_wwns = fc_targets
            fc_targets.extend(metro_fc_targets)
            for initiator in initiator_wwns:
                init_targ_map[initiator] = fc_targets

        return list(set(target_wwns)), init_targ_map

    def extend_volume(self, volume, new_size):
        """Extend an existing volume.

        :param volume: the cinder volume object
        :param new_size: the required new size
        """
        self.common.extend_volume(volume, new_size)

    def _update_volume_stats(self):
        """Retrieve stats info from volume group."""
        LOG.debug("Updating volume stats")
        data = self.common.update_volume_stats()
        data['storage_protocol'] = 'FC'
        data['driver_version'] = self.VERSION
        self._stats = data

    def manage_existing(self, volume, external_ref):
        """Manages an existing PowerMax/VMAX Volume (import to Cinder).

        Renames the Volume to match the expected name for the volume.
        Also need to consider things like QoS, Emulation, account/tenant.
        :param volume: the volume object
        :param external_ref: the reference for the PowerMax/VMAX volume
        :returns: model_update
        """
        return self.common.manage_existing(volume, external_ref)

    def manage_existing_get_size(self, volume, external_ref):
        """Return size of an existing PowerMax/VMAX volume to manage_existing.

        :param self: reference to class
        :param volume: the volume object including the volume_type_id
        :param external_ref: reference to the existing volume
        :returns: size of the volume in GB
        """
        return self.common.manage_existing_get_size(volume, external_ref)

    def unmanage(self, volume):
        """Export PowerMax/VMAX volume from Cinder.

        Leave the volume intact on the backend array.
        """
        return self.common.unmanage(volume)

    def manage_existing_snapshot(self, snapshot, existing_ref):
        """Manage an existing PowerMax/VMAX Snapshot (import to Cinder).

        Renames the Snapshot to prefix it with OS- to indicate
        it is managed by Cinder.

        :param snapshot: the snapshot object
        :param existing_ref: the snapshot name on the backend PowerMax/VMAX
        :returns: model_update
        """
        return self.common.manage_existing_snapshot(snapshot, existing_ref)

    def manage_existing_snapshot_get_size(self, snapshot, existing_ref):
        """Return the size of the source volume for manage-existing-snapshot.

        :param snapshot: the snapshot object
        :param existing_ref: the snapshot name on the backend PowerMax/VMAX
        :returns: size of the source volume in GB
        """
        return self.common.manage_existing_snapshot_get_size(snapshot)

    def unmanage_snapshot(self, snapshot):
        """Export PowerMax/VMAX Snapshot from Cinder.

        Leaves the snapshot intact on the backend PowerMax/VMAX.

        :param snapshot: the snapshot object
        """
        self.common.unmanage_snapshot(snapshot)

    def get_manageable_volumes(self, cinder_volumes, marker, limit, offset,
                               sort_keys, sort_dirs):
        """Lists all manageable volumes.

        :param cinder_volumes: List of currently managed Cinder volumes.
                               Unused in driver.
        :param marker: Begin returning volumes that appear later in the volume
                       list than that represented by this reference.
        :param limit: Maximum number of volumes to return. Default=1000.
        :param offset: Number of volumes to skip after marker.
        :param sort_keys: Results sort key. Valid keys: size, reference.
        :param sort_dirs: Results sort direction. Valid dirs: asc, desc.
        :returns: List of dicts containing all manageable volumes.
        """
        return self.common.get_manageable_volumes(marker, limit, offset,
                                                  sort_keys, sort_dirs)

    def get_manageable_snapshots(self, cinder_snapshots, marker, limit, offset,
                                 sort_keys, sort_dirs):
        """Lists all manageable snapshots.

        :param cinder_snapshots: List of currently managed Cinder snapshots.
                                 Unused in driver.
        :param marker: Begin returning volumes that appear later in the
                       snapshot list than that represented by this reference.
        :param limit: Maximum number of snapshots to return. Default=1000.
        :param offset: Number of snapshots to skip after marker.
        :param sort_keys: Results sort key. Valid keys: size, reference.
        :param sort_dirs: Results sort direction. Valid dirs: asc, desc.
        :returns: List of dicts containing all manageable snapshots.
        """
        return self.common.get_manageable_snapshots(marker, limit, offset,
                                                    sort_keys, sort_dirs)

    def retype(self, ctxt, volume, new_type, diff, host):
        """Migrate volume to another host using retype.

        :param ctxt: context
        :param volume: the volume object including the volume_type_id
        :param new_type: the new volume type.
        :param diff: difference between old and new volume types.
            Unused in driver.
        :param host: the host dict holding the relevant
            target(destination) information
        :returns: boolean -- True if retype succeeded, False if error
        """
        return self.common.retype(volume, new_type, host)

    def failover_host(self, context, volumes, secondary_id=None, groups=None):
        """Failover volumes to a secondary host/ backend.

        :param context: the context
        :param volumes: the list of volumes to be failed over
        :param secondary_id: the backend to be failed over to, is 'default'
                             if fail back
        :param groups: replication groups
        :returns: secondary_id, volume_update_list, group_update_list
        """
        return self.common.failover_host(volumes, secondary_id, groups)

    def create_group(self, context, group):
        """Creates a generic volume group.

        :param context: the context
        :param group: the group object
        :returns: model_update
        """
        return self.common.create_group(context, group)

    def delete_group(self, context, group, volumes):
        """Deletes a generic volume group.

        :param context: the context
        :param group: the group object
        :param volumes: the member volumes
        """
        return self.common.delete_group(
            context, group, volumes)

    def create_group_snapshot(self, context, group_snapshot, snapshots):
        """Creates a group snapshot.

        :param context: the context
        :param group_snapshot: the grouop snapshot
        :param snapshots: snapshots list
        """
        return self.common.create_group_snapshot(context,
                                                 group_snapshot, snapshots)

    def delete_group_snapshot(self, context, group_snapshot, snapshots):
        """Deletes a group snapshot.

        :param context: the context
        :param group_snapshot: the grouop snapshot
        :param snapshots: snapshots list
        """
        return self.common.delete_group_snapshot(context,
                                                 group_snapshot, snapshots)

    def update_group(self, context, group,
                     add_volumes=None, remove_volumes=None):
        """Updates LUNs in generic volume group.

        :param context: the context
        :param group: the group object
        :param add_volumes: flag for adding volumes
        :param remove_volumes: flag for removing volumes
        """
        return self.common.update_group(group, add_volumes,
                                        remove_volumes)

    def create_group_from_src(
            self, context, group, volumes, group_snapshot=None,
            snapshots=None, source_group=None, source_vols=None):
        """Creates the volume group from source.

        :param context: the context
        :param group: the group object to be created
        :param volumes: volumes in the group
        :param group_snapshot: the source volume group snapshot
        :param snapshots: snapshots of the source volumes
        :param source_group: the dictionary of a volume group as source.
        :param source_vols: a list of volume dictionaries in the source_group.
        """
        return self.common.create_group_from_src(
            context, group, volumes, group_snapshot, snapshots, source_group,
            source_vols)

    def enable_replication(self, context, group, volumes):
        """Enable replication for a group.

        :param context: the context
        :param group: the group object
        :param volumes: the list of volumes
        :returns: model_update, None
        """
        return self.common.enable_replication(context, group, volumes)

    def disable_replication(self, context, group, volumes):
        """Disable replication for a group.

        :param context: the context
        :param group: the group object
        :param volumes: the list of volumes
        :returns: model_update, None
        """
        return self.common.disable_replication(context, group, volumes)

    def failover_replication(self, context, group, volumes,
                             secondary_backend_id=None):
        """Failover replication for a group.

        :param context: the context
        :param group: the group object
        :param volumes: the list of volumes
        :param secondary_backend_id: the secondary backend id - default None
        :returns: model_update, vol_model_updates
        """
        return self.common.failover_replication(
            context, group, volumes, secondary_backend_id)

    def revert_to_snapshot(self, context, volume, snapshot):
        """Revert volume to snapshot

        :param context: the context
        :param volume: the cinder volume object
        :param snapshot: the cinder snapshot object
        """
        self.common.revert_to_snapshot(volume, snapshot)




# Copyright (c) 2013 eBay Inc.
# Copyright (c) 2013 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit Tests for qos specs internal API."""

import time
from unittest import mock

from oslo_db import exception as db_exc
from oslo_utils import timeutils
import six

from cinder import context
from cinder import db
from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.tests.unit import utils as test_utils
from cinder.volume import qos_specs
from cinder.volume import volume_types


def fake_db_qos_specs_create(context, values):
    if values['name'] == 'DupQoSName':
        raise exception.QoSSpecsExists(specs_id=values['name'])
    elif values['name'] == 'FailQoSName':
        raise db_exc.DBError()

    pass


def fake_db_get_vol_type(vol_type_number=1):
    return {'name': 'type-' + six.text_type(vol_type_number),
            'id': fake.QOS_SPEC_ID,
            'updated_at': None,
            'created_at': None,
            'deleted_at': None,
            'description': 'desc',
            'deleted': False,
            'is_public': True,
            'projects': [],
            'qos_specs_id': fake.QOS_SPEC_ID,
            'extra_specs': None}


class QoSSpecsTestCase(test.TestCase):
    """Test cases for qos specs code."""
    def setUp(self):
        super(QoSSpecsTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def _create_qos_specs(self, name, consumer='back-end', values=None):
        """Create a transfer object."""
        if values is None:
            values = {'key1': 'value1', 'key2': 'value2'}

        specs = {'name': name,
                 'consumer': consumer,
                 'specs': values}
        return db.qos_specs_create(self.ctxt, specs)['id']

    def test_create(self):
        input = {'key1': 'value1',
                 'key2': 'value2',
                 'key3': 'value3'}
        ref = qos_specs.create(self.ctxt, 'FakeName', input)
        specs_obj = qos_specs.get_qos_specs(self.ctxt, ref['id'])
        specs_obj_dic = {'consumer': specs_obj['consumer'],
                         'id': specs_obj['id'],
                         'name': specs_obj['name'],
                         'specs': specs_obj['specs']}
        expected = {'consumer': 'back-end',
                    'id': ref['id'],
                    'name': 'FakeName',
                    'specs': input}
        self.assertDictEqual(expected,
                             specs_obj_dic)

        # qos specs must have unique name
        self.assertRaises(exception.QoSSpecsExists,
                          qos_specs.create, self.ctxt, 'FakeName', input)

        # consumer must be one of: front-end, back-end, both
        input['consumer'] = 'fake'
        self.assertRaises(exception.InvalidQoSSpecs,
                          qos_specs.create, self.ctxt, 'QoSName', input)

        del input['consumer']

        self.mock_object(db, 'qos_specs_create',
                         fake_db_qos_specs_create)
        # able to catch DBError
        self.assertRaises(exception.QoSSpecsCreateFailed,
                          qos_specs.create, self.ctxt, 'FailQoSName', input)

    def test_update(self):
        def fake_db_update(context, specs_id, values):
            raise db_exc.DBError()

        qos = {'consumer': 'back-end',
               'specs': {'key1': 'value1'}}

        # qos specs must exists
        self.assertRaises(exception.QoSSpecsNotFound,
                          qos_specs.update, self.ctxt, 'fake_id', qos['specs'])

        specs_id = self._create_qos_specs('Name',
                                          qos['consumer'],
                                          qos['specs'])

        qos_specs.update(self.ctxt, specs_id,
                         {'key1': 'newvalue1', 'key2': 'value2'})

        specs = qos_specs.get_qos_specs(self.ctxt, specs_id)
        self.assertEqual('newvalue1', specs['specs']['key1'])
        self.assertEqual('value2', specs['specs']['key2'])

        # consumer must be one of: front-end, back-end, both
        self.assertRaises(exception.InvalidQoSSpecs,
                          qos_specs.update, self.ctxt, specs_id,
                          {'consumer': 'not-real'})

        self.mock_object(db, 'qos_specs_update', fake_db_update)
        self.assertRaises(exception.QoSSpecsUpdateFailed,
                          qos_specs.update, self.ctxt, specs_id, {'key':
                                                                  'new_key'})

    def test_delete(self):
        qos_id = self._create_qos_specs('my_qos')

        def fake_db_associations_get(context, id):
            vol_types = []
            if id == qos_id:
                vol_types = [fake_db_get_vol_type(id)]
            return vol_types

        def fake_db_delete(context, id):
            return {'deleted': True,
                    'deleted_at': timeutils.utcnow()}

        def fake_disassociate_all(context, id):
            pass

        self.mock_object(db, 'qos_specs_associations_get',
                         fake_db_associations_get)
        self.mock_object(qos_specs, 'disassociate_all',
                         fake_disassociate_all)
        self.mock_object(db, 'qos_specs_delete', fake_db_delete)
        self.assertRaises(exception.InvalidQoSSpecs,
                          qos_specs.delete, self.ctxt, None)
        self.assertRaises(exception.QoSSpecsNotFound,
                          qos_specs.delete, self.ctxt, 'NotFound')
        self.assertRaises(exception.QoSSpecsInUse,
                          qos_specs.delete, self.ctxt, qos_id)
        # able to delete in-use qos specs if force=True
        qos_specs.delete(self.ctxt, qos_id, force=True)

        # Can delete without forcing when no volume types
        qos_id_with_no_vol_types = self._create_qos_specs('no_vol_types')
        qos_specs.delete(self.ctxt, qos_id_with_no_vol_types, force=False)

    def test_delete_keys(self):
        def fake_db_qos_delete_key(context, id, key):
            if key == 'NotFound':
                raise exception.QoSSpecsKeyNotFound(specs_id=id,
                                                    specs_key=key)
            else:
                pass

        value = {'foo': 'Foo', 'bar': 'Bar', 'zoo': 'tiger'}
        name = 'QoSName'
        consumer = 'front-end'
        specs_id = self._create_qos_specs(name, consumer, value)
        qos_specs.delete_keys(self.ctxt, specs_id, ['foo', 'bar'])

        del value['foo']
        del value['bar']
        expected = {'name': name,
                    'id': specs_id,
                    'consumer': consumer,
                    'specs': value}
        specs = qos_specs.get_qos_specs(self.ctxt, specs_id)
        specs_dic = {'consumer': specs['consumer'],
                     'id': specs['id'],
                     'name': specs['name'],
                     'specs': specs['specs']}
        self.assertDictEqual(expected, specs_dic)

        self.mock_object(db, 'qos_specs_item_delete', fake_db_qos_delete_key)
        self.assertRaises(exception.InvalidQoSSpecs,
                          qos_specs.delete_keys, self.ctxt, None, [])
        self.assertRaises(exception.QoSSpecsNotFound,
                          qos_specs.delete_keys, self.ctxt, 'NotFound', [])
        self.assertRaises(exception.QoSSpecsKeyNotFound,
                          qos_specs.delete_keys, self.ctxt,
                          specs_id, ['NotFound'])
        self.assertRaises(exception.QoSSpecsKeyNotFound,
                          qos_specs.delete_keys, self.ctxt, specs_id,
                          ['foo', 'bar', 'NotFound'])

    @mock.patch.object(db, 'qos_specs_associations_get')
    def test_get_associations(self, mock_qos_specs_associations_get):
        vol_types = [fake_db_get_vol_type(x) for x in range(2)]

        mock_qos_specs_associations_get.return_value = vol_types
        specs_id = self._create_qos_specs('new_spec')
        res = qos_specs.get_associations(self.ctxt, specs_id)
        for vol_type in vol_types:
            expected_type = {
                'association_type': 'volume_type',
                'id': vol_type['id'],
                'name': vol_type['name']
            }
            self.assertIn(expected_type, res)

        e = exception.QoSSpecsNotFound(specs_id='Trouble')
        mock_qos_specs_associations_get.side_effect = e
        self.assertRaises(exception.CinderException,
                          qos_specs.get_associations, self.ctxt,
                          'Trouble')

    def test_associate_qos_with_type(self):
        def fake_qos_specs_get(context, id):
            if id == 'NotFound':
                raise exception.QoSSpecsNotFound(specs_id=id)
            else:
                pass

        def fake_db_associate(context, id, type_id):
            if id == 'Trouble':
                raise db_exc.DBError()
            elif type_id == 'NotFound':
                raise exception.VolumeTypeNotFound(volume_type_id=type_id)
            pass

        def fake_vol_type_qos_get(type_id):
            if type_id == 'Invalid':
                return {'qos_specs': {'id': 'Invalid'}}
            else:
                return {'qos_specs': None}

        type_ref = volume_types.create(self.ctxt, 'TypeName')
        specs_id = self._create_qos_specs('QoSName')

        qos_specs.associate_qos_with_type(self.ctxt, specs_id,
                                          type_ref['id'])
        res = qos_specs.get_associations(self.ctxt, specs_id)
        self.assertEqual(1, len(res))
        self.assertEqual('TypeName', res[0]['name'])
        self.assertEqual(type_ref['id'], res[0]['id'])

        self.mock_object(db, 'qos_specs_associate',
                         fake_db_associate)
        self.mock_object(qos_specs, 'get_qos_specs', fake_qos_specs_get)
        self.mock_object(volume_types, 'get_volume_type_qos_specs',
                         fake_vol_type_qos_get)
        self.assertRaises(exception.VolumeTypeNotFound,
                          qos_specs.associate_qos_with_type,
                          self.ctxt, 'specs-id', 'NotFound')
        self.assertRaises(exception.QoSSpecsAssociateFailed,
                          qos_specs.associate_qos_with_type,
                          self.ctxt, 'Trouble', 'id')
        self.assertRaises(exception.QoSSpecsNotFound,
                          qos_specs.associate_qos_with_type,
                          self.ctxt, 'NotFound', 'id')
        self.assertRaises(exception.InvalidVolumeType,
                          qos_specs.associate_qos_with_type,
                          self.ctxt, 'specs-id', 'Invalid')

    def test_disassociate_qos_specs(self):
        def fake_db_disassociate(context, id, type_id):
            raise db_exc.DBError()

        type_ref = volume_types.create(self.ctxt, 'TypeName')
        specs_id = self._create_qos_specs('QoSName')

        qos_specs.associate_qos_with_type(self.ctxt, specs_id,
                                          type_ref['id'])
        res = qos_specs.get_associations(self.ctxt, specs_id)
        self.assertEqual(1, len(res))

        qos_specs.disassociate_qos_specs(self.ctxt, specs_id, type_ref['id'])
        res = qos_specs.get_associations(self.ctxt, specs_id)
        self.assertEqual(0, len(res))

        self.assertRaises(exception.VolumeTypeNotFound,
                          qos_specs.disassociate_qos_specs,
                          self.ctxt, specs_id, 'NotFound')

        # Verify we can disassociate specs from volume_type even if they are
        # not associated with no error
        qos_specs.disassociate_qos_specs(self.ctxt, specs_id, type_ref['id'])
        qos_specs.associate_qos_with_type(self.ctxt, specs_id, type_ref['id'])
        self.mock_object(db, 'qos_specs_disassociate',
                         fake_db_disassociate)
        self.assertRaises(exception.QoSSpecsDisassociateFailed,
                          qos_specs.disassociate_qos_specs,
                          self.ctxt, specs_id, type_ref['id'])

    def test_disassociate_all(self):
        def fake_db_disassociate_all(context, id):
            if id == 'Trouble':
                raise db_exc.DBError()
            pass

        def fake_qos_specs_get(context, id):
            if id == 'NotFound':
                raise exception.QoSSpecsNotFound(specs_id=id)
            else:
                pass

        type1_ref = volume_types.create(self.ctxt, 'TypeName1')
        type2_ref = volume_types.create(self.ctxt, 'TypeName2')
        specs_id = self._create_qos_specs('QoSName')

        qos_specs.associate_qos_with_type(self.ctxt, specs_id,
                                          type1_ref['id'])
        qos_specs.associate_qos_with_type(self.ctxt, specs_id,
                                          type2_ref['id'])
        res = qos_specs.get_associations(self.ctxt, specs_id)
        self.assertEqual(2, len(res))

        qos_specs.disassociate_all(self.ctxt, specs_id)
        res = qos_specs.get_associations(self.ctxt, specs_id)
        self.assertEqual(0, len(res))

        self.mock_object(db, 'qos_specs_disassociate_all',
                         fake_db_disassociate_all)
        self.mock_object(qos_specs, 'get_qos_specs',
                         fake_qos_specs_get)
        self.assertRaises(exception.QoSSpecsDisassociateFailed,
                          qos_specs.disassociate_all,
                          self.ctxt, 'Trouble')

    def test_get_all_specs(self):
        qos_specs_list = [{'name': 'Specs1',
                           'created_at': None,
                           'updated_at': None,
                           'deleted_at': None,
                           'deleted': None,
                           'consumer': 'both',
                           'specs': {'key1': 'value1',
                                     'key2': 'value2',
                                     'key3': 'value3'}},
                          {'name': 'Specs2',
                           'created_at': None,
                           'updated_at': None,
                           'deleted_at': None,
                           'deleted': None,
                           'consumer': 'both',
                           'specs': {'key1': 'value1',
                                     'key2': 'value2',
                                     'key3': 'value3',
                                     'key4': 'value4'}}]

        for index, qos_specs_dict in enumerate(qos_specs_list):
            qos_specs_id = self._create_qos_specs(
                qos_specs_dict['name'],
                qos_specs_dict['consumer'],
                qos_specs_dict['specs'])
            qos_specs_dict['id'] = qos_specs_id
            specs = db.qos_specs_get(self.ctxt, qos_specs_id)
            qos_specs_list[index]['created_at'] = test_utils.time_format(
                specs['created_at'])

        res = qos_specs.get_all_specs(self.ctxt)
        self.assertEqual(len(qos_specs_list), len(res))

        qos_res_simple_dict = []
        # Need to make list of dictionaries instead of VOs for assertIn to work
        for qos in res:
            qos_res_simple_dict.append(
                qos.obj_to_primitive()['versioned_object.data'])
        for qos_spec in qos_specs_list:
            self.assertIn(qos_spec, qos_res_simple_dict)

    def test_get_qos_specs(self):
        one_time_value = str(int(time.time()))
        specs = {'key1': one_time_value,
                 'key2': 'value2',
                 'key3': 'value3'}
        qos_id = self._create_qos_specs('Specs1', 'both', specs)
        specs = qos_specs.get_qos_specs(self.ctxt, qos_id)
        self.assertEqual(one_time_value, specs['specs']['key1'])
        self.assertRaises(exception.InvalidQoSSpecs,
                          qos_specs.get_qos_specs, self.ctxt, None)



# Copyright (C) 2013 eBay Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import common


class ViewBuilder(common.ViewBuilder):
    """Model QoS specs API responses as a python dictionary."""

    _collection_name = "qos-specs"

    def __init__(self):
        """Initialize view builder."""
        super(ViewBuilder, self).__init__()

    def summary_list(self, request, qos_specs, qos_count=None):
        """Show a list of qos_specs without many details."""
        return self._list_view(self.detail, request, qos_specs, qos_count)

    def summary(self, request, qos_spec):
        """Generic, non-detailed view of a qos_specs."""
        return self.detail(request, qos_spec)

    def detail(self, request, qos_spec):
        """Detailed view of a single qos_spec."""
        # TODO(zhiteng) Add associations to detailed view
        return {
            'qos_specs': {
                'id': qos_spec.id,
                'name': qos_spec.name,
                'consumer': qos_spec.consumer,
                'specs': qos_spec.specs,
            },
            'links': self._get_links(request,
                                     qos_spec.id),
        }

    def associations(self, request, associates):
        """View of qos specs associations."""
        return {
            'qos_associations': associates
        }

    def _list_view(self, func, request, qos_specs, qos_count=None):
        """Provide a view for a list of qos_specs."""
        specs_list = [func(request, specs)['qos_specs'] for specs in qos_specs]
        specs_links = self._get_collection_links(request, qos_specs,
                                                 self._collection_name,
                                                 qos_count)
        specs_dict = dict(qos_specs=specs_list)
        if specs_links:
            specs_dict['qos_specs_links'] = specs_links

        return specs_dict



# Copyright 2011 OpenStack Foundation
# Copyright 2011 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
WSGI middleware for OpenStack Volume API.
"""

from cinder.api import extensions
import cinder.api.openstack
from cinder.api.v2 import limits
from cinder.api.v2 import snapshot_metadata
from cinder.api.v2 import snapshots
from cinder.api.v2 import types
from cinder.api.v2 import volume_metadata
from cinder.api.v2 import volumes
from cinder.api import versions


class APIRouter(cinder.api.openstack.APIRouter):
    """Routes requests on the API to the appropriate controller and method."""
    ExtensionManager = extensions.ExtensionManager

    def _setup_routes(self, mapper, ext_mgr):
        self.resources['versions'] = versions.create_resource()
        mapper.connect("versions", "/",
                       controller=self.resources['versions'],
                       action='index')

        mapper.redirect("", "/")

        self.resources['volumes'] = volumes.create_resource(ext_mgr)
        mapper.resource("volume", "volumes",
                        controller=self.resources['volumes'],
                        collection={'detail': 'GET'},
                        member={'action': 'POST'})

        self.resources['types'] = types.create_resource()
        mapper.resource("type", "types",
                        controller=self.resources['types'],
                        member={'action': 'POST'})

        self.resources['snapshots'] = snapshots.create_resource(ext_mgr)
        mapper.resource("snapshot", "snapshots",
                        controller=self.resources['snapshots'],
                        collection={'detail': 'GET'},
                        member={'action': 'POST'})

        self.resources['limits'] = limits.create_resource()
        mapper.resource("limit", "limits",
                        controller=self.resources['limits'])

        self.resources['snapshot_metadata'] = \
            snapshot_metadata.create_resource()
        snapshot_metadata_controller = self.resources['snapshot_metadata']

        mapper.resource("snapshot_metadata", "metadata",
                        controller=snapshot_metadata_controller,
                        parent_resource=dict(member_name='snapshot',
                                             collection_name='snapshots'))

        mapper.connect("metadata",
                       "/{project_id}/snapshots/{snapshot_id}/metadata",
                       controller=snapshot_metadata_controller,
                       action='update_all',
                       conditions={"method": ['PUT']})

        self.resources['volume_metadata'] = volume_metadata.create_resource()
        volume_metadata_controller = self.resources['volume_metadata']

        mapper.resource("volume_metadata", "metadata",
                        controller=volume_metadata_controller,
                        parent_resource=dict(member_name='volume',
                                             collection_name='volumes'))

        mapper.connect("metadata",
                       "/{project_id}/volumes/{volume_id}/metadata",
                       controller=volume_metadata_controller,
                       action='update_all',
                       conditions={"method": ['PUT']})



#!/usr/bin/env python

# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Starter script for Cinder Volume Backup."""

import logging as python_logging
import shlex
import sys

# NOTE: Monkey patching must go before OSLO.log import, otherwise OSLO.context
# will not use greenthread thread local and all greenthreads will share the
# same context.  It's also a good idea to monkey patch everything before
# loading multiprocessing
import eventlet
eventlet.monkey_patch()
# Monkey patch the original current_thread to use the up-to-date _active
# global variable. See https://bugs.launchpad.net/bugs/1863021 and
# https://github.com/eventlet/eventlet/issues/592
import __original_module_threading as orig_threading
import threading # noqa
orig_threading.current_thread.__globals__['_active'] = threading._active

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_privsep import priv_context
from oslo_reports import guru_meditation_report as gmr
from oslo_reports import opts as gmr_opts

# Need to register global_opts
from cinder.common import config  # noqa
from cinder.db import api as session
from cinder import i18n
i18n.enable_lazy()
from cinder import objects
from cinder import service
from cinder import utils
from cinder import version


CONF = cfg.CONF

backup_cmd_opts = [
    cfg.IntOpt('backup_workers',
               default=1, min=1, max=processutils.get_worker_count(),
               help='Number of backup processes to launch. '
               'Improves performance with concurrent backups.'),
    cfg.IntOpt('backup_max_operations',
               default=15,
               min=0,
               help='Maximum number of concurrent memory heavy operations: '
                    'backup and restore. Value of 0 means unlimited'),
]
CONF.register_opts(backup_cmd_opts)

LOG = None

# NOTE(mriedem): The default backup driver uses swift and performs read/write
# operations in a thread. swiftclient will log requests and responses at DEBUG
# level, which can cause a thread switch and break the backup operation. So we
# set a default log level of WARN for swiftclient to try and avoid this issue.
_EXTRA_DEFAULT_LOG_LEVELS = ['swiftclient=WARN']


def _launch_backup_process(launcher, num_process, _semaphore):
    try:
        server = service.Service.create(binary='cinder-backup',
                                        coordination=True,
                                        service_name='backup',
                                        process_number=num_process,
                                        semaphore=_semaphore)
    except Exception:
        LOG.exception('Backup service %s failed to start.', CONF.host)
        sys.exit(1)
    else:
        # Dispose of the whole DB connection pool here before
        # starting another process.  Otherwise we run into cases where
        # child processes share DB connections which results in errors.
        session.dispose_engine()
        launcher.launch_service(server)


def main():
    objects.register_all()
    gmr_opts.set_defaults(CONF)
    CONF(sys.argv[1:], project='cinder',
         version=version.version_string())
    logging.set_defaults(
        default_log_levels=logging.get_default_log_levels() +
        _EXTRA_DEFAULT_LOG_LEVELS)
    logging.setup(CONF, "cinder")
    python_logging.captureWarnings(True)
    priv_context.init(root_helper=shlex.split(utils.get_root_helper()))
    utils.monkey_patch()
    gmr.TextGuruMeditation.setup_autorun(version, conf=CONF)
    global LOG
    LOG = logging.getLogger(__name__)
    semaphore = utils.semaphore_factory(CONF.backup_max_operations,
                                        CONF.backup_workers)

    LOG.info('Backup running with %s processes.', CONF.backup_workers)
    launcher = service.get_launcher()

    for i in range(1, CONF.backup_workers + 1):
        _launch_backup_process(launcher, i, semaphore)

    launcher.wait()



# Copyright (c) 2015 Infortrend Technology, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Infortrend basic CLI factory.
"""

import abc
import os
import time

from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import strutils
import six

from cinder import utils

LOG = logging.getLogger(__name__)

DEFAULT_RETRY_TIME = 5


def retry_cli(func):
    def inner(self, *args, **kwargs):
        total_retry_time = self.cli_retry_time

        if total_retry_time is None:
            total_retry_time = DEFAULT_RETRY_TIME

        retry_time = 0
        while retry_time < total_retry_time:
            rc, out = func(self, *args, **kwargs)
            retry_time += 1

            if rc == 0:
                break

            LOG.error(
                'Retry %(retry)s times: %(method)s Failed '
                '%(rc)s: %(reason)s', {
                    'retry': retry_time,
                    'method': self.__class__.__name__,
                    'rc': rc,
                    'reason': out})

            # show error log, not retrying
            if rc == 1:
                # RAID return fail
                break
            elif rc == 11:
                # rc == 11 means not exist
                break
            elif rc == 20:
                # rc == 20 means already exist
                break

        LOG.debug(
            'Method: %(method)s Return Code: %(rc)s '
            'Output: %(out)s', {
                'method': self.__class__.__name__, 'rc': rc, 'out': out})
        return rc, out
    return inner


def os_execute(fd, raidcmd_timeout, command_line):
    os.write(fd, command_line.encode('utf-8'))
    return os_read(fd, 8192, 'RAIDCmd:>', raidcmd_timeout)


def os_read(fd, buffer_size, cmd_pattern, raidcmd_timeout):
    content = ''
    start_time = int(time.time())
    while True:
        time.sleep(0.5)
        output = os.read(fd, buffer_size)
        if len(output) > 0:
            content += output.decode('utf-8')
        if content.find(cmd_pattern) >= 0:
            break
        if int(time.time()) - start_time > raidcmd_timeout:
            content = 'Raidcmd timeout: %s' % content
            LOG.error(
                'Raidcmd exceeds cli timeout [%(timeout)s]s.', {
                    'timeout': raidcmd_timeout})
            break
    return content


def strip_empty_in_list(list):
    result = []
    for entry in list:
        entry = entry.strip()
        if entry != "":
            result.append(entry)

    return result


def table_to_dict(table):
    tableHeader = table[0].split("  ")
    tableHeaderList = strip_empty_in_list(tableHeader)

    result = []

    for i in range(len(table) - 2):
        if table[i + 2].strip() == "":
            break

        resultEntry = {}
        tableEntry = table[i + 2].split("  ")
        tableEntryList = strip_empty_in_list(tableEntry)

        for key, value in zip(tableHeaderList, tableEntryList):
            resultEntry[key] = value

        result.append(resultEntry)
    return result


def content_lines_to_dict(content_lines):
    result = []
    resultEntry = {}

    for content_line in content_lines:

        if content_line.strip() == "":
            result.append(resultEntry)
            resultEntry = {}
            continue

        split_entry = content_line.strip().split(": ", 1)
        resultEntry[split_entry[0]] = split_entry[1]

    return result


@six.add_metaclass(abc.ABCMeta)
class BaseCommand(object):

    """The BaseCommand abstract class."""

    def __init__(self):
        super(BaseCommand, self).__init__()

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        pass


class ShellCommand(BaseCommand):

    """The Common ShellCommand."""

    def __init__(self, cli_conf):
        super(ShellCommand, self).__init__()
        self.cli_retry_time = cli_conf.get('cli_retry_time')

    @retry_cli
    def execute(self, *args, **kwargs):
        commands = ' '.join(args)
        result = None
        rc = 0
        try:
            result, err = utils.execute(commands, shell=True)
        except processutils.ProcessExecutionError as pe:
            rc = pe.exit_code
            result = pe.stdout
            result = result.replace('\n', '\\n')
            LOG.error(
                'Error on execute command. '
                'Error code: %(exit_code)d Error msg: %(result)s', {
                    'exit_code': pe.exit_code, 'result': result})
        return rc, result


class ExecuteCommand(BaseCommand):

    """The Cinder Filter Command."""

    def __init__(self, cli_conf):
        super(ExecuteCommand, self).__init__()
        self.cli_retry_time = cli_conf.get('cli_retry_time')

    @retry_cli
    def execute(self, *args, **kwargs):
        result = None
        rc = 0
        try:
            result, err = utils.execute(*args, **kwargs)
        except processutils.ProcessExecutionError as pe:
            rc = pe.exit_code
            result = pe.stdout
            result = result.replace('\n', '\\n')
            LOG.error(
                'Error on execute command. '
                'Error code: %(exit_code)d Error msg: %(result)s', {
                    'exit_code': pe.exit_code, 'result': result})
        return rc, result


class CLIBaseCommand(BaseCommand):

    """The CLIBaseCommand class."""

    def __init__(self, cli_conf):
        super(CLIBaseCommand, self).__init__()
        self.cli_retry_time = cli_conf.get('cli_retry_time')
        self.raidcmd_timeout = cli_conf.get('raidcmd_timeout')
        self.cli_cache = cli_conf.get('cli_cache')
        self.pid = cli_conf.get('pid')
        self.fd = cli_conf.get('fd')
        self.command = ""
        self.parameters = ()
        self.show_noinit = ""
        self.command_line = ""

    def _generate_command(self, parameters):
        """Generate execute Command. use java, execute, command, parameters."""
        self.parameters = parameters
        parameters_line = ' '.join(parameters)

        self.command_line = "{0} {1} {2}\n".format(
            self.command,
            parameters_line,
            self.show_noinit)

        return self.command_line

    def _parser(self, content=None):
        """The parser to parse command result.

        :param content: The parse Content
        :returns: parse result
        """
        content = content.replace("\r", "")
        content = content.replace("\\/-", "")
        content = content.strip()
        LOG.debug(content)

        if content is not None:
            content_lines = content.split("\n")
            rc, out = self._parse_return(content_lines)

            if rc != 0:
                return rc, out
            else:
                return rc, content_lines

        return -1, None

    @retry_cli
    def execute(self, *args, **kwargs):
        command_line = self._generate_command(args)
        LOG.debug('Executing: %(command)s', {
            'command': strutils.mask_password(command_line)})
        rc = 0
        result = None
        try:
            content = self._execute(command_line)
            rc, result = self._parser(content)
        except processutils.ProcessExecutionError as pe:
            rc = -2  # prevent confusing with cli real rc
            result = pe.stdout
            result = result.replace('\n', '\\n')
            LOG.error(
                'Error on execute %(command)s. '
                'Error code: %(exit_code)d Error msg: %(result)s', {
                    'command': strutils.mask_password(command_line),
                    'exit_code': pe.exit_code,
                    'result': result})
        return rc, result

    def _execute(self, command_line):
        return os_execute(
            self.fd, self.raidcmd_timeout, command_line)

    def _parse_return(self, content_lines):
        """Get the end of command line result."""
        rc = 0
        if 'Raidcmd timeout' in content_lines[0]:
            rc = -3
            return_cli_result = content_lines
        elif len(content_lines) < 4:
            rc = -4
            return_cli_result = 'Raidcmd output error: %s' % content_lines
        else:
            return_value = content_lines[-3].strip().split(' ', 1)[1]
            return_cli_result = content_lines[-4].strip().split(' ', 1)[1]
            rc = int(return_value, 16)

        return rc, return_cli_result


class ConnectRaid(CLIBaseCommand):

    """The Connect Raid Command."""

    def __init__(self, *args, **kwargs):
        super(ConnectRaid, self).__init__(*args, **kwargs)
        self.command = "connect"


class CheckConnection(CLIBaseCommand):

    """The Check Connection Command."""

    def __init__(self, *args, **kwargs):
        super(CheckConnection, self).__init__(*args, **kwargs)
        self.command = "lock"


class InitCache(CLIBaseCommand):
    """Refresh cacahe data for update volume status."""

    def __init__(self, *args, **kwargs):
        super(InitCache, self).__init__(*args, **kwargs)
        self.command = "utility init-cache"


class CreateLD(CLIBaseCommand):

    """The Create LD Command."""

    def __init__(self, *args, **kwargs):
        super(CreateLD, self).__init__(*args, **kwargs)
        self.command = "create ld"


class CreateLV(CLIBaseCommand):

    """The Create LV Command."""

    def __init__(self, *args, **kwargs):
        super(CreateLV, self).__init__(*args, **kwargs)
        self.command = "create lv"


class CreatePartition(CLIBaseCommand):

    """Create Partition.

    create part
        [LV-ID] [name] [size={partition-size}]
        [min={minimal-reserve-size}] [init={switch}]
        [tier={tier-level-list}]
    """

    def __init__(self, *args, **kwargs):
        super(CreatePartition, self).__init__(*args, **kwargs)
        self.command = "create part"


class DeletePartition(CLIBaseCommand):

    """Delete Partition.

    delete part [partition-ID] [-y]
    """

    def __init__(self, *args, **kwargs):
        super(DeletePartition, self).__init__(*args, **kwargs)
        self.command = "delete part"


class SetPartition(CLIBaseCommand):

    """Set Partition.

    set part
    [partition-ID] [name={partition-name}] [min={minimal-reserve-size}]
    set part expand [partition-ID] [size={expand-size}]
    set part purge [partition-ID] [number] [rule-type]
    set part reclaim [partition-ID]
    set part tier-resided [partition-ID] tier={tier-level-list}
    """

    def __init__(self, *args, **kwargs):
        super(SetPartition, self).__init__(*args, **kwargs)
        self.command = "set part"


class SetLV(CLIBaseCommand):

    """Set Logical Volume.

    set lv tier-migrate [LV-ID] [part={partition-IDs}]
    """

    def __init__(self, *args, **kwargs):
        super(SetLV, self).__init__(*args, **kwargs)
        self.command = "set lv"


class SetSnapshot(CLIBaseCommand):

    """Set Logical Volume.

    set lv tier-migrate [LV-ID] [part={partition-IDs}]
    """

    def __init__(self, *args, **kwargs):
        super(SetSnapshot, self).__init__(*args, **kwargs)
        self.command = "set si"


class CreateMap(CLIBaseCommand):

    """Map the Partition on the channel.

    create map
        [part] [partition-ID] [Channel-ID]
        [Target-ID] [LUN-ID] [assign={assign-to}]
    """

    def __init__(self, *args, **kwargs):
        super(CreateMap, self).__init__(*args, **kwargs)
        self.command = "create map"


class DeleteMap(CLIBaseCommand):

    """Unmap the Partition on the channel.

    delete map
        [part] [partition-ID] [Channel-ID]
        [Target-ID] [LUN-ID] [-y]
    """

    def __init__(self, *args, **kwargs):
        super(DeleteMap, self).__init__(*args, **kwargs)
        self.command = "delete map"


class CreateSnapshot(CLIBaseCommand):

    """Create partition's Snapshot.

    create si [part] [partition-ID]
    """

    def __init__(self, *args, **kwargs):
        super(CreateSnapshot, self).__init__(*args, **kwargs)
        self.command = "create si"


class DeleteSnapshot(CLIBaseCommand):

    """Delete partition's Snapshot.

    delete si [snapshot-image-ID] [-y]
    """

    def __init__(self, *args, **kwargs):
        super(DeleteSnapshot, self).__init__(*args, **kwargs)
        self.command = "delete si"


class CreateReplica(CLIBaseCommand):

    """Create partition or snapshot's replica.

    create replica
        [name] [part | si] [source-volume-ID]
        [part] [target-volume-ID] [type={replication-mode}]
        [priority={level}] [desc={description}]
        [incremental={switch}] [timeout={value}]
        [compression={switch}]
    """

    def __init__(self, *args, **kwargs):
        super(CreateReplica, self).__init__(*args, **kwargs)
        self.command = "create replica"


class DeleteReplica(CLIBaseCommand):

    """Delete and terminate specific replication job.

    delete replica [volume-pair-ID] [-y]
    """

    def __init__(self, *args, **kwargs):
        super(DeleteReplica, self).__init__(*args, **kwargs)
        self.command = "delete replica"


class CreateIQN(CLIBaseCommand):

    """Create host iqn for CHAP or lun filter.

    create iqn
        [IQN] [IQN-alias-name] [user={username}] [password={secret}]
        [target={name}] [target-password={secret}] [ip={ip-address}]
        [mask={netmask-ip}]
    """

    def __init__(self, *args, **kwargs):
        super(CreateIQN, self).__init__(*args, **kwargs)
        self.command = "create iqn"


class DeleteIQN(CLIBaseCommand):

    """Delete host iqn by name.

    delete iqn [name]
    """

    def __init__(self, *args, **kwargs):
        super(DeleteIQN, self).__init__(*args, **kwargs)
        self.command = "delete iqn"


class SetIOTimeout(CLIBaseCommand):

    """Set CLI IO timeout.

    utility set io-timeout [time]
    """

    def __init__(self, *args, **kwargs):
        super(SetIOTimeout, self).__init__(*args, **kwargs)
        self.command = "utility set io-timeout"


class ShowCommand(CLIBaseCommand):

    """Basic Show Command."""

    def __init__(self, *args, **kwargs):
        super(ShowCommand, self).__init__(*args, **kwargs)
        self.param_detail = "-l"
        self.default_type = "table"
        self.start_key = ""
        if self.cli_cache:
            self.show_noinit = "-noinit"

    def _parser(self, content=None):
        """Parse Table or Detail format into dict.

        # Table format

         ID   Name  LD-amount
        ----------------------
         123  LV-1  1

        # Result

        {
            'ID': '123',
            'Name': 'LV-1',
            'LD-amount': '1'
        }

        # Detail format

         ID: 5DE94FF775D81C30
         Name: LV-1
         LD-amount: 1

        # Result

        {
            'ID': '123',
            'Name': 'LV-1',
            'LD-amount': '1'
        }

        :param content: The parse Content.
        :returns: parse result
        """
        rc, out = super(ShowCommand, self)._parser(content)

        # Error.
        if rc != 0:
            return rc, out

        # No content.
        if len(out) < 6:
            return rc, []

        detect_type = self.detect_type()

        # Show detail content.
        if detect_type == "list":

            start_id = self.detect_detail_start_index(out)

            if start_id < 0:
                return rc, []

            result = content_lines_to_dict(out[start_id:-3])
        else:

            start_id = self.detect_table_start_index(out)

            if start_id < 0:
                return rc, []

            result = table_to_dict(out[start_id:-4])

        return rc, result

    def detect_type(self):
        if self.param_detail in self.parameters:
            detect_type = "list"
        else:
            detect_type = self.default_type
        return detect_type

    def detect_table_start_index(self, content):
        for i in range(1, len(content)):
            key = content[i].strip().split('  ')
            if self.start_key in key[0].strip():
                return i

        return -1

    def detect_detail_start_index(self, content):
        for i in range(1, len(content)):
            split_entry = content[i].strip().split(' ')
            if len(split_entry) >= 2 and ':' in split_entry[0]:
                return i

        return -1


class ShowLD(ShowCommand):

    """Show LD.

    show ld [index-list]
    """

    def __init__(self, *args, **kwargs):
        super(ShowLD, self).__init__(*args, **kwargs)
        self.command = "show ld"


class ShowLV(ShowCommand):

    """Show LV.

    show lv [lv={LV-IDs}] [-l]
    """

    def __init__(self, *args, **kwargs):
        super(ShowLV, self).__init__(*args, **kwargs)
        self.command = "show lv"
        self.start_key = "ID"
        self.show_noinit = ""

    def detect_table_start_index(self, content):
        if "tier" in self.parameters:
            self.start_key = "LV-Name"

        for i in range(1, len(content)):
            key = content[i].strip().split('  ')
            if self.start_key in key[0].strip():
                return i

        return -1


class ShowPartition(ShowCommand):

    """Show Partition.

    show part [part={partition-IDs} | lv={LV-IDs}] [-l]
    """

    def __init__(self, *args, **kwargs):
        super(ShowPartition, self).__init__(*args, **kwargs)
        self.command = "show part"
        self.start_key = "ID"
        self.show_noinit = ""


class ShowSnapshot(ShowCommand):

    """Show Snapshot.

    show si [si={snapshot-image-IDs} | part={partition-IDs} | lv={LV-IDs}] [-l]
    """

    def __init__(self, *args, **kwargs):
        super(ShowSnapshot, self).__init__(*args, **kwargs)
        self.command = "show si"
        self.start_key = "Index"


class ShowDevice(ShowCommand):

    """Show Device.

    show device
    """

    def __init__(self, *args, **kwargs):
        super(ShowDevice, self).__init__(*args, **kwargs)
        self.command = "show device"
        self.start_key = "Index"


class ShowChannel(ShowCommand):

    """Show Channel.

    show channel
    """

    def __init__(self, *args, **kwargs):
        super(ShowChannel, self).__init__(*args, **kwargs)
        self.command = "show channel"
        self.start_key = "Ch"


class ShowDisk(ShowCommand):

    """The Show Disk Command.

    show disk [disk-index-list | channel={ch}]
    """

    def __init__(self, *args, **kwargs):
        super(ShowDisk, self).__init__(*args, **kwargs)
        self.command = "show disk"


class ShowMap(ShowCommand):

    """Show Map.

    show map [part={partition-IDs} | channel={channel-IDs}] [-l]
    """

    def __init__(self, *args, **kwargs):
        super(ShowMap, self).__init__(*args, **kwargs)
        self.command = "show map"
        self.start_key = "Ch"


class ShowNet(ShowCommand):

    """Show IP network.

    show net [id={channel-IDs}] [-l]
    """

    def __init__(self, *args, **kwargs):
        super(ShowNet, self).__init__(*args, **kwargs)
        self.command = "show net"
        self.start_key = "ID"


class ShowLicense(ShowCommand):

    """Show License.

    show license
    """

    def __init__(self, *args, **kwargs):
        super(ShowLicense, self).__init__(*args, **kwargs)
        self.command = "show license"
        self.start_key = "License"

    def _parser(self, content=None):
        """Parse License format.

        # License format

         License  Amount(Partition/Subsystem)  Expired
        ------------------------------------------------
         EonPath  ---                          True

        # Result

        {
            'EonPath': {
                'Amount': '---',
                'Support': True
            }
        }

        :param content: The parse Content.
        :returns: parse result
        """
        rc, out = super(ShowLicense, self)._parser(content)

        if rc != 0:
            return rc, out

        if len(out) > 0:
            result = {}
            for entry in out:
                if entry['Expired'] == '---' or entry['Expired'] == 'Expired':
                    support = False
                else:
                    support = True
                result[entry['License']] = {
                    'Amount':
                        entry['Amount(Partition/Subsystem)'],
                    'Support': support
                }
            return rc, result

        return rc, []


class ShowReplica(ShowCommand):

    """Show information of all replication jobs or specific job.

    show replica [id={volume-pair-IDs}] [-l] id={volume-pair-IDs}
    """

    def __init__(self, *args, **kwargs):
        super(ShowReplica, self).__init__(*args, **kwargs)
        self.command = 'show replica'
        self.show_noinit = ""


class ShowWWN(ShowCommand):

    """Show Fibre network.

    show wwn
    """

    def __init__(self, *args, **kwargs):
        super(ShowWWN, self).__init__(*args, **kwargs)
        self.command = "show wwn"
        self.start_key = "CH"


class ShowIQN(ShowCommand):

    """Show iSCSI initiator IQN which is set by create iqn.

    show iqn
    """

    LIST_START_LINE = "List of initiator IQN(s):"

    def __init__(self, *args, **kwargs):
        super(ShowIQN, self).__init__(*args, **kwargs)
        self.command = "show iqn"
        self.default_type = "list"

    def detect_detail_start_index(self, content):
        for i in range(1, len(content)):
            if content[i].strip() == self.LIST_START_LINE:
                return i + 2

        return -1


class ShowHost(ShowCommand):

    """Show host settings.

    show host
    """

    def __init__(self, *args, **kwargs):
        super(ShowHost, self).__init__(*args, **kwargs)
        self.command = "show host"
        self.default_type = "list"

    def detect_detail_start_index(self, content):
        for i in range(1, len(content)):
            if ':' in content[i]:
                return i
        return -1



#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.policies import volumes as policy


class VolumeTenantAttributeController(wsgi.Controller):
    def _add_volume_tenant_attribute(self, req, resp_volume):
        db_volume = req.get_db_volume(resp_volume['id'])
        key = "%s:tenant_id" % Volume_tenant_attribute.alias
        resp_volume[key] = db_volume['project_id']

    @wsgi.extends
    def show(self, req, resp_obj, id):
        context = req.environ['cinder.context']
        if context.authorize(policy.TENANT_ATTRIBUTE_POLICY, fatal=False):
            volume = resp_obj.obj['volume']
            self._add_volume_tenant_attribute(req, volume)

    @wsgi.extends
    def detail(self, req, resp_obj):
        context = req.environ['cinder.context']
        if context.authorize(policy.TENANT_ATTRIBUTE_POLICY, fatal=False):
            for vol in list(resp_obj.obj['volumes']):
                self._add_volume_tenant_attribute(req, vol)


class Volume_tenant_attribute(extensions.ExtensionDescriptor):
    """Expose the internal project_id as an attribute of a volume."""

    name = "VolumeTenantAttribute"
    alias = "os-vol-tenant-attr"
    updated = "2011-11-03T00:00:00+00:00"

    def get_controller_extensions(self):
        controller = VolumeTenantAttributeController()
        extension = extensions.ControllerExtension(self, 'volumes', controller)
        return [extension]



#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""The Volume Image Metadata API extension."""
from oslo_log import log as logging
from six.moves import http_client
import webob

from cinder.api import common
from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_image_metadata
from cinder.api import validation
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.policies import volume_metadata as policy
from cinder import volume

LOG = logging.getLogger(__name__)


class VolumeImageMetadataController(wsgi.Controller):
    def __init__(self, *args, **kwargs):
        super(VolumeImageMetadataController, self).__init__(*args, **kwargs)
        self.volume_api = volume.API()

    def _get_image_metadata(self, context, volume_id):
        # Not found exception will be handled at the wsgi level
        volume = self.volume_api.get(context, volume_id)
        meta = self.volume_api.get_volume_image_metadata(context, volume)
        return (volume, meta)

    def _add_image_metadata(self, context, resp_volume_list, image_metas=None):
        """Appends the image metadata to each of the given volume.

        :param context: the request context
        :param resp_volume_list: the response volume list
        :param image_metas: The image metadata to append, if None is provided
                            it will be retrieved from the database. An empty
                            dict means there is no metadata and it should not
                            be retrieved from the db.
        """
        vol_id_list = []
        for vol in resp_volume_list:
            vol_id_list.append(vol['id'])
        if image_metas is None:
            try:
                image_metas = self.volume_api.get_list_volumes_image_metadata(
                    context, vol_id_list)
            except Exception as e:
                LOG.debug('Get image metadata error: %s', e)
                return
        if image_metas:
            for vol in resp_volume_list:
                image_meta = image_metas.get(vol['id'])
                if image_meta:
                    vol['volume_image_metadata'] = dict(image_meta)

    @wsgi.extends
    def show(self, req, resp_obj, id):
        context = req.environ['cinder.context']
        if context.authorize(policy.IMAGE_METADATA_POLICY, fatal=False):
            self._add_image_metadata(context, [resp_obj.obj['volume']])

    @wsgi.extends
    def detail(self, req, resp_obj):
        context = req.environ['cinder.context']
        if context.authorize(policy.IMAGE_METADATA_POLICY, fatal=False):
            # Just get the image metadata of those volumes in response.
            volumes = list(resp_obj.obj.get('volumes', []))
            if volumes:
                self._add_image_metadata(context, volumes)

    @wsgi.action("os-set_image_metadata")
    @validation.schema(volume_image_metadata.set_image_metadata)
    def create(self, req, id, body):
        context = req.environ['cinder.context']
        volume = objects.Volume.get_by_id(context, id)
        if context.authorize(policy.IMAGE_METADATA_POLICY,
                             target_obj=volume):
            metadata = body['os-set_image_metadata']['metadata']
            new_metadata = self._update_volume_image_metadata(context,
                                                              id,
                                                              metadata,
                                                              delete=False)

            return {'metadata': new_metadata}

    def _update_volume_image_metadata(self, context,
                                      volume_id,
                                      metadata,
                                      delete=False):
        try:
            volume = self.volume_api.get(context, volume_id)
            return self.volume_api.update_volume_metadata(
                context,
                volume,
                metadata,
                delete=False,
                meta_type=common.METADATA_TYPES.image)
        # Not found exception will be handled at the wsgi level
        except (ValueError, AttributeError):
            msg = _("Malformed request body.")
            raise webob.exc.HTTPBadRequest(explanation=msg)
        except exception.InvalidVolumeMetadata as error:
            raise webob.exc.HTTPBadRequest(explanation=error.msg)
        except exception.InvalidVolumeMetadataSize as error:
            raise webob.exc.HTTPRequestEntityTooLarge(explanation=error.msg)

    @wsgi.action("os-show_image_metadata")
    def index(self, req, id, body):
        context = req.environ['cinder.context']
        return {'metadata': self._get_image_metadata(context, id)[1]}

    @wsgi.action("os-unset_image_metadata")
    @validation.schema(volume_image_metadata.unset_image_metadata)
    def delete(self, req, id, body):
        """Deletes an existing image metadata."""
        context = req.environ['cinder.context']
        volume = objects.Volume.get_by_id(context, id)
        if context.authorize(policy.IMAGE_METADATA_POLICY, target_obj=volume):
            key = body['os-unset_image_metadata']['key']

            vol, metadata = self._get_image_metadata(context, id)
            if key not in metadata:
                raise exception.GlanceMetadataNotFound(id=id)

            self.volume_api.delete_volume_metadata(
                context, vol, key,
                meta_type=common.METADATA_TYPES.image)

            return webob.Response(status_int=http_client.OK)


class Volume_image_metadata(extensions.ExtensionDescriptor):
    """Show image metadata associated with the volume."""

    name = "VolumeImageMetadata"
    alias = "os-vol-image-meta"
    updated = "2012-12-07T00:00:00+00:00"

    def get_controller_extensions(self):
        controller = VolumeImageMetadataController()
        extension = extensions.ControllerExtension(self, 'volumes', controller)
        return [extension]



#    (c) Copyright 2014 Cisco Systems Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""
Script to push the zone configuration to Cisco SAN switches.
"""
import random
import re

from eventlet import greenthread
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import excutils
import six

from cinder import exception
from cinder.i18n import _
from cinder import ssh_utils
from cinder import utils
from cinder.zonemanager.drivers.cisco import exception as c_exception
import cinder.zonemanager.drivers.cisco.fc_zone_constants as ZoneConstant

LOG = logging.getLogger(__name__)


class CiscoFCZoneClientCLI(object):
    """Cisco FC zone client cli implementation.

    OpenStack Fibre Channel zone client cli connector
    to manage FC zoning in Cisco SAN fabrics.

    Version history:
        1.0 - Initial Cisco FC zone client cli
    """

    switch_ip = None
    switch_port = '22'
    switch_user = 'admin'
    switch_pwd = 'none'

    def __init__(self, ipaddress, username, password, port, vsan):
        """initializing the client."""
        self.switch_ip = ipaddress
        self.switch_port = port
        self.switch_user = username
        self.switch_pwd = password
        self.fabric_vsan = vsan
        self.sshpool = None

    def get_active_zone_set(self):
        """Return the active zone configuration.

        Return active zoneset from fabric. When none of the configurations
        are active then it will return empty map.

        :returns: Map -- active zone set map in the following format

        .. code-block:: python

            {
                'zones':
                    {'openstack50060b0000c26604201900051ee8e329':
                        ['50060b0000c26604', '201900051ee8e329']
                    },
                'active_zone_config': 'OpenStack_Cfg'
            }
        """
        zone_set = {}
        zone = {}
        zone_member = None
        zone_name = None
        switch_data = None
        zone_set_name = None
        try:
            switch_data = self._get_switch_info(
                [ZoneConstant.GET_ACTIVE_ZONE_CFG, self.fabric_vsan,
                 ' | no-more'])
        except c_exception.CiscoZoningCliException:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed getting active zone set "
                          "from fabric %s", self.switch_ip)
        try:
            for line in switch_data:
                # Split on non-word characters,
                line_split = re.split(r'[\s\[\]]+', line)
                if ZoneConstant.CFG_ZONESET in line_split:
                    # zoneset name [name] vsan [vsan]
                    zone_set_name = \
                        line_split[line_split.index(ZoneConstant.CFG_ZONESET)
                                   + 2]
                    continue
                if ZoneConstant.CFG_ZONE in line_split:
                    # zone name [name] vsan [vsan]
                    zone_name = \
                        line_split[line_split.index(ZoneConstant.CFG_ZONE) + 2]
                    zone[zone_name] = list()
                    continue
                if ZoneConstant.CFG_ZONE_MEMBER in line_split:
                    # Examples:
                    # pwwn c0:50:76:05:15:9f:00:12
                    # * fcid 0x1e01c0 [pwwn 50:05:07:68:02:20:48:04] [V7K_N1P2]
                    zone_member = \
                        line_split[
                            line_split.index(ZoneConstant.CFG_ZONE_MEMBER) + 1]
                    zone_member_list = zone.get(zone_name)
                    zone_member_list.append(zone_member)

            zone_set[ZoneConstant.CFG_ZONES] = zone
            zone_set[ZoneConstant.ACTIVE_ZONE_CONFIG] = zone_set_name
        except Exception as ex:
            # In case of parsing error here, it should be malformed cli output.
            msg = _("Malformed zone configuration: (switch=%(switch)s "
                    "zone_config=%(zone_config)s)."
                    ) % {'switch': self.switch_ip,
                         'zone_config': switch_data}
            LOG.error(msg)
            exc_msg = _("Exception: %s") % six.text_type(ex)
            LOG.error(exc_msg)
            raise exception.FCZoneDriverException(reason=msg)

        return zone_set

    def add_zones(self, zones, activate, fabric_vsan, active_zone_set,
                  zone_status):
        """Add zone configuration.

        This method will add the zone configuration passed by user.

        :param zones: Zone names mapped to members and VSANs
                      Zone members are colon separated but case-insensitive

        .. code-block:: python

            {   zonename1:[zonememeber1,zonemember2,...],
                zonename2:[zonemember1, zonemember2,...]...}

            e.g:

            {
                'openstack50060b0000c26604201900051ee8e329':
                  ['50:06:0b:00:00:c2:66:04', '20:19:00:05:1e:e8:e3:29']
            }

        :param activate: True will activate the zone config.
        :param fabric_vsan:
        :param active_zone_set: Active zone set dict retrieved from
                                get_active_zone_set method
        :param zone_status: Status of the zone
        :raises CiscoZoningCliException:
        """
        LOG.debug("Add Zones - Zones passed: %s", zones)

        LOG.debug("Active zone set: %s", active_zone_set)
        zone_list = active_zone_set[ZoneConstant.CFG_ZONES]
        LOG.debug("zone list: %s", zone_list)
        LOG.debug("zone status: %s", zone_status)

        cfg_name = active_zone_set[ZoneConstant.ACTIVE_ZONE_CONFIG]

        zone_cmds = [['conf'],
                     ['zoneset', 'name', cfg_name, 'vsan', fabric_vsan]]

        for zone in zones.keys():
            zone_cmds.append(['zone', 'name', zone])

            for member in zones[zone]:
                zone_cmds.append(['member', 'pwwn', member])

        zone_cmds.append(['end'])

        try:
            LOG.debug("Add zones: Config cmd to run: %s", zone_cmds)
            self._ssh_execute(zone_cmds, True, 1)

            if activate:
                self.activate_zoneset(cfg_name, fabric_vsan, zone_status)
            self._cfg_save()
        except Exception as e:

            msg = _("Creating and activating zone set failed: "
                    "(Zone set=%(zoneset)s error=%(err)s)."
                    ) % {'zoneset': cfg_name, 'err': six.text_type(e)}
            LOG.error(msg)
            raise c_exception.CiscoZoningCliException(reason=msg)

    def update_zones(self, zones, activate, fabric_vsan, operation,
                     active_zone_set, zone_status):
        """Update the zone configuration.

        This method will update the zone configuration passed by user.

        :param zones: zone names mapped to members. Zone members
                      are colon separated but case-insensitive

        .. code-block:: python

            {   zonename1:[zonememeber1, zonemember2,...],
                zonename2:[zonemember1, zonemember2,...]...}

            e.g:

            {
                'openstack50060b0000c26604201900051ee8e329':
                        ['50:06:0b:00:00:c2:66:04',
                         '20:19:00:05:1e:e8:e3:29']
            }

        :param activate: True will activate the zone config.
        :param operation: zone add or zone remove
        :param fabric_vsan: Virtual San #
        :param active_zone_set: Active zone set dict retrieved from
                                get_active_zone_set method
        :param zone_status: Status of the zone
        :raises CiscoZoningCliException:
        """

        LOG.debug("Update Zones - Operation: %(op)s - Zones "
                  "passed: %(zones)s",
                  {'op': operation, 'zones': zones})

        cfg_name = active_zone_set[ZoneConstant.ACTIVE_ZONE_CONFIG]

        zone_cmds = [['conf'],
                     ['zoneset', 'name', cfg_name, 'vsan', fabric_vsan]]
        zone_mod_cmd = []
        if operation == ZoneConstant.ZONE_ADD:
            zone_mod_cmd = ['member', 'pwwn']
        elif operation == ZoneConstant.ZONE_REMOVE:
            zone_mod_cmd = ['no', 'member', 'pwwn']

        for zone, zone_members in zones.items():
            zone_cmds.append(['zone', 'name', zone])
            for member in zone_members:
                zone_cmds.append(zone_mod_cmd + [member])
        zone_cmds.append(['end'])

        try:
            LOG.debug("Update zones: Config cmd to run: %s", zone_cmds)
            self._ssh_execute(zone_cmds, True, 1)

            if activate:
                self.activate_zoneset(cfg_name, fabric_vsan, zone_status)
            self._cfg_save()
        except Exception as e:

            msg = (_("Updating and activating zone set failed: "
                     "(Zone set=%(zoneset)s error=%(err)s).")
                   % {'zoneset': cfg_name, 'err': six.text_type(e)})
            LOG.error(msg)
            raise c_exception.CiscoZoningCliException(reason=msg)

    def activate_zoneset(self, cfgname, fabric_vsan, zone_status):
        """Method to Activate the zone config. Param cfgname - ZonesetName."""

        LOG.debug("zone status: %s", zone_status)

        cmd_list = [['conf'],
                    ['zoneset', 'activate', 'name', cfgname, 'vsan',
                     self.fabric_vsan]]
        if zone_status['mode'] == 'enhanced':
            cmd_list.append(['zone', 'commit', 'vsan', fabric_vsan])

        cmd_list.append(['end'])

        return self._ssh_execute(cmd_list, True, 1)

    def get_zoning_status(self):
        """Return the zoning mode and session for a zoneset."""
        zone_status = {}

        try:
            switch_data = self._get_switch_info(
                [ZoneConstant.GET_ZONE_STATUS, self.fabric_vsan])
        except c_exception.CiscoZoningCliException:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed getting zone status "
                          "from fabric %s", self.switch_ip)
        try:
            for line in switch_data:
                # Split on non-word characters,
                line_split = re.split(r'[\s\[\]]+', line)
                if 'mode:' in line_split:
                    # mode: <enhanced|basic>
                    zone_status['mode'] = line_split[line_split.index('mode:')
                                                     + 1]
                    continue
                if 'session:' in line_split:
                    # session: <none|a value other than none>
                    zone_status['session'] = \
                        line_split[line_split.index('session:') + 1]
                    continue
        except Exception as ex:
            # In case of parsing error here, it should be malformed cli output.
            msg = _("Malformed zone status: (switch=%(switch)s "
                    "zone_config=%(zone_config)s)."
                    ) % {'switch': self.switch_ip,
                         'zone_status': switch_data}
            LOG.error(msg)
            exc_msg = _("Exception: %s") % six.text_type(ex)
            LOG.error(exc_msg)
            raise exception.FCZoneDriverException(reason=msg)

        return zone_status

    def delete_zones(self, zone_names, activate, fabric_vsan, active_zone_set,
                     zone_status):
        """Delete zones from fabric.

        Method to delete the active zone config zones

        params zone_names: zoneNames separated by semicolon
        params activate: True/False
        """

        LOG.debug("zone_names %s", zone_names)
        active_zoneset_name = active_zone_set[ZoneConstant.ACTIVE_ZONE_CONFIG]

        cmds = [['conf'],
                ['zoneset', 'name', active_zoneset_name, 'vsan',
                 fabric_vsan]]

        try:
            for zone in set(zone_names.split(';')):
                cmds.append(['no', 'zone', 'name', zone])

            cmds.append(['end'])

            LOG.debug("Delete zones: Config cmd to run: %s", cmds)
            self._ssh_execute(cmds, True, 1)

            if activate:
                self.activate_zoneset(active_zoneset_name, fabric_vsan,
                                      zone_status)
            self._cfg_save()

        except Exception as e:
            msg = _("Deleting zones failed: (command=%(cmd)s error=%(err)s)."
                    ) % {'cmd': cmds, 'err': six.text_type(e)}
            LOG.error(msg)
            raise c_exception.CiscoZoningCliException(reason=msg)

    def get_nameserver_info(self):
        """Get name server data from fabric.

        This method will return the connected node port wwn list(local
        and remote) for the given switch fabric

        show fcns database
        """
        cli_output = None
        return_list = []
        try:
            cli_output = self._get_switch_info([ZoneConstant.FCNS_SHOW,
                                                self.fabric_vsan])
        except c_exception.CiscoZoningCliException:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed collecting fcns database "
                          "info for fabric %s", self.switch_ip)

        if (cli_output):
            return_list = self._parse_ns_output(cli_output)

        LOG.info("Connector returning fcnsinfo-%s", return_list)

        return return_list

    @utils.retry(processutils.ProcessExecutionError, retries=5)
    def _cfg_save(self):
        cmd = ['copy', 'running-config', 'startup-config']
        self._run_ssh(cmd, True)

    def _get_switch_info(self, cmd_list):
        stdout, stderr, sw_data = None, None, None
        try:
            stdout, stderr = self._run_ssh(cmd_list, True)
            LOG.debug("CLI output from ssh - output: %s", stdout)
            if (stdout):
                sw_data = stdout.splitlines()
            return sw_data
        except processutils.ProcessExecutionError as e:
            msg = _("Error while getting data via ssh: (command=%(cmd)s "
                    "error=%(err)s).") % {'cmd': cmd_list,
                                          'err': six.text_type(e)}
            LOG.error(msg)
            raise c_exception.CiscoZoningCliException(reason=msg)

    def _parse_ns_output(self, switch_data):
        """Parses name server data.

        Parses nameserver raw data and adds the device port wwns to the list

        :returns: List -- list of device port wwn from ns info
        """
        return_list = []
        for line in switch_data:
            if not(" N " in line):
                continue
            linesplit = line.split()
            if len(linesplit) > 2:
                node_port_wwn = linesplit[2]
                return_list.append(node_port_wwn)
            else:
                msg = _("Malformed show fcns database string: %s") % line
                LOG.error(msg)
                raise exception.InvalidParameterValue(err=msg)
        return return_list

    def _run_ssh(self, cmd_list, check_exit_code=True):

        command = ' '.join(cmd_list)

        if not self.sshpool:
            self.sshpool = ssh_utils.SSHPool(self.switch_ip,
                                             self.switch_port,
                                             None,
                                             self.switch_user,
                                             self.switch_pwd,
                                             min_size=1,
                                             max_size=5)
        try:
            with self.sshpool.item() as ssh:
                return processutils.ssh_execute(
                    ssh,
                    command,
                    check_exit_code=check_exit_code)

        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.warning("Error running SSH command: %s", command)

    def _ssh_execute(self, cmd_list, check_exit_code=True, attempts=1):
        """Execute cli with status update.

        Executes CLI commands where status return is expected.

        cmd_list is a list of commands, where each command is itself
        a list of parameters.  We use utils.check_ssh_injection to check each
        command, but then join then with " ; " to form a single command.
        """

        # Check that each command is secure
        for cmd in cmd_list:
            utils.check_ssh_injection(cmd)

        # Combine into a single command.
        command = ' ; '.join(map(lambda x: ' '.join(x), cmd_list))

        if not self.sshpool:
            self.sshpool = ssh_utils.SSHPool(self.switch_ip,
                                             self.switch_port,
                                             None,
                                             self.switch_user,
                                             self.switch_pwd,
                                             min_size=1,
                                             max_size=5)
        stdin, stdout, stderr = None, None, None
        LOG.debug("Executing command via ssh: %s", command)
        last_exception = None
        try:
            with self.sshpool.item() as ssh:
                while attempts > 0:
                    attempts -= 1
                    try:
                        stdin, stdout, stderr = ssh.exec_command(command)
                        channel = stdout.channel
                        exit_status = channel.recv_exit_status()
                        LOG.debug("Exit Status from ssh: %s", exit_status)
                        # exit_status == -1 if no exit code was returned
                        if exit_status != -1:
                            LOG.debug('Result was %s', exit_status)
                            if check_exit_code and exit_status != 0:
                                raise processutils.ProcessExecutionError(
                                    exit_code=exit_status,
                                    stdout=stdout,
                                    stderr=stderr,
                                    cmd=command)
                            else:
                                return True
                        else:
                            return True
                    except Exception as e:
                        LOG.exception('Error executing SSH command.')
                        last_exception = e
                        greenthread.sleep(random.randint(20, 500) / 100.0)
                LOG.debug("Handling error case after SSH: %s", last_exception)
                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1,
                        stdout="",
                        stderr="Error running SSH command",
                        cmd=command)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.exception("Error executing command via ssh.")
        finally:
            if stdin:
                stdin.flush()
                stdin.close()
            if stdout:
                stdout.close()
            if stderr:
                stderr.close()

    def cleanup(self):
        self.sshpool = None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_concurrency import processutils
from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume.targets import driver
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)


class ISCSITarget(driver.Target):
    """Target object for block storage devices.

    Base class for target object, where target
    is data transport mechanism (target) specific calls.
    This includes things like create targets, attach, detach
    etc.
    """

    def __init__(self, *args, **kwargs):
        super(ISCSITarget, self).__init__(*args, **kwargs)
        self.iscsi_target_prefix = self.configuration.safe_get('target_prefix')
        self.iscsi_protocol = self.configuration.safe_get('target_protocol')
        self.protocol = 'iSCSI'
        self.volumes_dir = self.configuration.safe_get('volumes_dir')

    def _get_iscsi_properties(self, volume, multipath=False):
        """Gets iscsi configuration

        We ideally get saved information in the volume entity, but fall back
        to discovery if need be. Discovery may be completely removed in the
        future.

        The properties are:

        :target_discovered:    boolean indicating whether discovery was used

        :target_iqn:    the IQN of the iSCSI target

        :target_portal:    the portal of the iSCSI target

        :target_lun:    the lun of the iSCSI target

        :volume_id:    the uuid of the volume

        :auth_method:, :auth_username:, :auth_password:

            the authentication details. Right now, either auth_method is not
            present meaning no authentication, or auth_method == `CHAP`
            meaning use CHAP with the specified credentials.

        :discard:    boolean indicating if discard is supported

        In some of drivers that support multiple connections (for multipath
        and for single path with failover on connection failure), it returns
        :target_iqns, :target_portals, :target_luns, which contain lists of
        multiple values. The main portal information is also returned in
        :target_iqn, :target_portal, :target_lun for backward compatibility.

        Note that some of drivers don't return :target_portals even if they
        support multipath. Then the connector should use sendtargets discovery
        to find the other portals if it supports multipath.
        """

        properties = {}

        location = volume['provider_location']

        if location:
            # provider_location is the same format as iSCSI discovery output
            properties['target_discovered'] = False
        else:
            location = self._do_iscsi_discovery(volume)

            if not location:
                msg = (_("Could not find iSCSI export for volume %s") %
                        (volume['name']))
                raise exception.InvalidVolume(reason=msg)

            LOG.debug("ISCSI Discovery: Found %s", location)
            properties['target_discovered'] = True

        results = location.split(" ")
        portals = results[0].split(",")[0].split(";")
        iqn = results[1]
        nr_portals = len(portals)
        try:
            lun = int(results[2])
        except (IndexError, ValueError):
            lun = 0

        if nr_portals > 1 or multipath:
            properties['target_portals'] = portals
            properties['target_iqns'] = [iqn] * nr_portals
            properties['target_luns'] = [lun] * nr_portals
        properties['target_portal'] = portals[0]
        properties['target_iqn'] = iqn
        properties['target_lun'] = lun

        properties['volume_id'] = volume['id']

        auth = volume['provider_auth']
        if auth:
            (auth_method, auth_username, auth_secret) = auth.split()

            properties['auth_method'] = auth_method
            properties['auth_username'] = auth_username
            properties['auth_password'] = auth_secret

        geometry = volume.get('provider_geometry', None)
        if geometry:
            (physical_block_size, logical_block_size) = geometry.split()
            properties['physical_block_size'] = physical_block_size
            properties['logical_block_size'] = logical_block_size

        encryption_key_id = volume.get('encryption_key_id', None)
        properties['encrypted'] = encryption_key_id is not None

        return properties

    def _iscsi_authentication(self, chap, name, password):
        return "%s %s %s" % (chap, name, password)

    def _do_iscsi_discovery(self, volume):
        # TODO(justinsb): Deprecate discovery and use stored info
        # NOTE(justinsb): Discovery won't work with CHAP-secured targets (?)
        LOG.warning("ISCSI provider_location not stored, using discovery")

        volume_id = volume['id']

        try:
            # NOTE(griff) We're doing the split straight away which should be
            # safe since using '@' in hostname is considered invalid

            (out, _err) = utils.execute('iscsiadm', '-m', 'discovery',
                                        '-t', 'sendtargets', '-p',
                                        volume['host'].split('@')[0],
                                        run_as_root=True)
        except processutils.ProcessExecutionError as ex:
            LOG.error("ISCSI discovery attempt failed for: %s",
                      volume['host'].split('@')[0])
            LOG.debug("Error from iscsiadm -m discovery: %s", ex.stderr)
            return None

        for target in out.splitlines():
            if (self.configuration.safe_get('target_ip_address') in target
                    and volume_id in target):
                return target
        return None

    def _get_portals_config(self):
        # Prepare portals configuration
        portals_ips = ([self.configuration.target_ip_address]
                       + self.configuration.iscsi_secondary_ip_addresses or [])

        return {'portals_ips': portals_ips,
                'portals_port': self.configuration.target_port}

    def create_export(self, context, volume, volume_path):
        """Creates an export for a logical volume."""
        # 'iscsi_name': 'iqn.2010-10.org.openstack:volume-00000001'
        iscsi_name = "%s%s" % (self.configuration.target_prefix,
                               volume['name'])
        iscsi_target, lun = self._get_target_and_lun(context, volume)

        # Verify we haven't setup a CHAP creds file already
        # if DNE no big deal, we'll just create it
        chap_auth = self._get_target_chap_auth(context, volume)
        if not chap_auth:
            chap_auth = (volume_utils.generate_username(),
                         volume_utils.generate_password())

        # Get portals ips and port
        portals_config = self._get_portals_config()

        # NOTE(jdg): For TgtAdm case iscsi_name is the ONLY param we need
        # should clean this all up at some point in the future
        tid = self.create_iscsi_target(iscsi_name,
                                       iscsi_target,
                                       lun,
                                       volume_path,
                                       chap_auth,
                                       **portals_config)
        data = {}
        data['location'] = self._iscsi_location(
            self.configuration.target_ip_address, tid, iscsi_name, lun,
            self.configuration.iscsi_secondary_ip_addresses)
        LOG.debug('Set provider_location to: %s', data['location'])
        data['auth'] = self._iscsi_authentication(
            'CHAP', *chap_auth)
        return data

    def remove_export(self, context, volume):
        try:
            iscsi_target, lun = self._get_target_and_lun(context, volume)
        except exception.NotFound:
            LOG.info("Skipping remove_export. No iscsi_target "
                     "provisioned for volume: %s", volume['id'])
            return
        try:

            # NOTE: provider_location may be unset if the volume hasn't
            # been exported
            location = volume['provider_location'].split(' ')
            iqn = location[1]

            # ietadm show will exit with an error
            # this export has already been removed
            self.show_target(iscsi_target, iqn=iqn)

        except Exception:
            LOG.info("Skipping remove_export. No iscsi_target "
                     "is presently exported for volume: %s", volume['id'])
            return

        # NOTE: For TgtAdm case volume['id'] is the ONLY param we need
        self.remove_iscsi_target(iscsi_target, lun, volume['id'],
                                 volume['name'])

    def ensure_export(self, context, volume, volume_path):
        """Recreates an export for a logical volume."""
        iscsi_name = "%s%s" % (self.configuration.target_prefix,
                               volume['name'])

        chap_auth = self._get_target_chap_auth(context, volume)

        # Get portals ips and port
        portals_config = self._get_portals_config()

        iscsi_target, lun = self._get_target_and_lun(context, volume)
        self.create_iscsi_target(
            iscsi_name, iscsi_target, lun, volume_path,
            chap_auth, check_exit_code=False,
            old_name=None, **portals_config)

    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection info.

        The iscsi driver returns a driver_volume_type of 'iscsi'.
        The format of the driver data is defined in _get_iscsi_properties.
        Example return value::

            {
                'driver_volume_type': 'iscsi'
                'data': {
                    'target_discovered': True,
                    'target_iqn': 'iqn.2010-10.org.openstack:volume-00000001',
                    'target_portal': '127.0.0.0.1:3260',
                    'volume_id': '9a0d35d0-175a-11e4-8c21-0800200c9a66',
                    'discard': False,
                }
            }
        """

        iscsi_properties = self._get_iscsi_properties(volume,
                                                      connector.get(
                                                          'multipath'))
        return {
            'driver_volume_type': self.iscsi_protocol,
            'data': iscsi_properties
        }

    def terminate_connection(self, volume, connector, **kwargs):
        pass

    def validate_connector(self, connector):
        # NOTE(jdg): api passes in connector which is initiator info
        if 'initiator' not in connector:
            err_msg = ('The volume driver requires the iSCSI initiator '
                       'name in the connector.')
            LOG.error(err_msg)
            raise exception.InvalidConnectorException(missing='initiator')
        return True

    def _iscsi_location(self, ip, target, iqn, lun=None, ip_secondary=None):
        ip_secondary = ip_secondary or []
        port = self.configuration.target_port
        portals = map(lambda x: "%s:%s" % (volume_utils.sanitize_host(x),
                                           port),
                      [ip] + ip_secondary)
        return ("%(portals)s,%(target)s %(iqn)s %(lun)s"
                % ({'portals': ";".join(portals),
                    'target': target, 'iqn': iqn, 'lun': lun}))

    def show_target(self, iscsi_target, iqn, **kwargs):
        if iqn is None:
            raise exception.InvalidParameterValue(
                err=_('valid iqn needed for show_target'))

        tid = self._get_target(iqn)
        if tid is None:
            raise exception.NotFound()

    def _get_target_chap_auth(self, context, volume):
        """Get the current chap auth username and password."""
        try:
            # Query DB to get latest state of volume
            volume_info = self.db.volume_get(context, volume['id'])
            # 'provider_auth': 'CHAP user_id password'
            if volume_info['provider_auth']:
                return tuple(volume_info['provider_auth'].split(' ', 3)[1:])
        except exception.NotFound:
            LOG.debug('Failed to get CHAP auth from DB for %s.', volume['id'])

    def extend_target(self, volume):
        """Reinitializes a target after the LV has been extended.

        Note: This will cause IO disruption in most cases.
        """
        iscsi_name = "%s%s" % (self.configuration.target_prefix,
                               volume['name'])

        if volume.volume_attachment:
            self._do_tgt_update(iscsi_name, force=True)

    @abc.abstractmethod
    def _get_target_and_lun(self, context, volume):
        """Get iscsi target and lun."""
        pass

    @abc.abstractmethod
    def create_iscsi_target(self, name, tid, lun, path,
                            chap_auth, **kwargs):
        pass

    @abc.abstractmethod
    def remove_iscsi_target(self, tid, lun, vol_id, vol_name, **kwargs):
        pass

    @abc.abstractmethod
    def _get_iscsi_target(self, context, vol_id):
        pass

    @abc.abstractmethod
    def _get_target(self, iqn):
        pass

    def _do_tgt_update(self, name, force=False):
        pass


class SanISCSITarget(ISCSITarget):
    """iSCSI target for san devices.

    San devices are slightly different, they don't need to implement
    all of the same things that we need to implement locally fro LVM
    and local block devices when we create and manage our own targets.

    """
    @abc.abstractmethod
    def create_export(self, context, volume, volume_path):
        pass

    @abc.abstractmethod
    def remove_export(self, context, volume):
        pass

    @abc.abstractmethod
    def ensure_export(self, context, volume, volume_path):
        pass

    @abc.abstractmethod
    def terminate_connection(self, volume, connector, **kwargs):
        pass

    # NOTE(jdg): Items needed for local iSCSI target drivers,
    # but NOT sans Stub them out here to make abc happy

    # Use care when looking at these to make sure something
    # that's inheritted isn't dependent on one of
    # these.
    def _get_target_and_lun(self, context, volume):
        pass

    def _get_target_chap_auth(self, context, volume):
        pass

    def create_iscsi_target(self, name, tid, lun, path,
                            chap_auth, **kwargs):
        pass

    def remove_iscsi_target(self, tid, lun, vol_id, vol_name, **kwargs):
        pass

    def _get_iscsi_target(self, context, vol_id):
        pass

    def _get_target(self, iqn):
        pass



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from oslo_log import log as logging

from cinder import exception
from cinder.i18n import _
from cinder import utils
from cinder.volume.drivers.huawei import constants

LOG = logging.getLogger(__name__)


class SmartQos(object):
    def __init__(self, client):
        self.client = client

    def _check_qos_consistency(self, policy, qos):
        for key in [k.upper() for k in constants.QOS_SPEC_KEYS]:
            if qos.get(key, '0') != policy.get(key, '0'):
                return False
        return True

    def _change_lun_priority(self, qos, lun_id):
        for key in qos:
            if key.startswith('MIN') or key.startswith('LATENCY'):
                data = {"IOPRIORITY": "3"}
                self.client.update_lun(lun_id, data)
                break

    @utils.synchronized('huawei_qos', external=True)
    def add(self, qos, lun_id):
        self._change_lun_priority(qos, lun_id)
        qos_id = self.client.create_qos(qos, lun_id)
        try:
            self.client.activate_deactivate_qos(qos_id, True)
        except exception.VolumeBackendAPIException:
            self.remove(qos_id, lun_id)
            raise

        return qos_id

    @utils.synchronized('huawei_qos', external=True)
    def remove(self, qos_id, lun_id, qos_info=None):
        if not qos_info:
            qos_info = self.client.get_qos_info(qos_id)
        lun_list = json.loads(qos_info['LUNLIST'])
        if lun_id in lun_list:
            lun_list.remove(lun_id)

        if len(lun_list) == 0:
            if qos_info['RUNNINGSTATUS'] != constants.QOS_INACTIVATED:
                self.client.activate_deactivate_qos(qos_id, False)
            self.client.delete_qos(qos_id)
        else:
            self.client.update_qos_luns(qos_id, lun_list)

    def update(self, qos_id, new_qos, lun_id):
        qos_info = self.client.get_qos_info(qos_id)
        if self._check_qos_consistency(qos_info, new_qos):
            return

        self.remove(qos_id, lun_id, qos_info)
        self.add(new_qos, lun_id)


class SmartPartition(object):
    def __init__(self, client):
        self.client = client

    def add(self, partitionname, lun_id):
        partition_id = self.client.get_partition_id_by_name(partitionname)
        if not partition_id:
            msg = _('Cannot find partition by name %s.') % partitionname
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        self.client.add_lun_to_partition(lun_id, partition_id)
        return partition_id

    def remove(self, partition_id, lun_id):
        self.client.remove_lun_from_partition(lun_id, partition_id)

    def update(self, partition_id, partitionname, lun_id):
        partition_info = self.client.get_partition_info_by_id(partition_id)
        if partition_info['NAME'] == partitionname:
            return

        self.remove(partition_id, lun_id)
        self.add(partitionname, lun_id)

    def check_partition_valid(self, partitionname):
        partition_id = self.client.get_partition_id_by_name(partitionname)
        if not partition_id:
            msg = _("Partition %s doesn't exist.") % partitionname
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)


class SmartCache(object):
    def __init__(self, client):
        self.client = client

    def add(self, cachename, lun_id):
        cache_id = self.client.get_cache_id_by_name(cachename)
        if not cache_id:
            msg = _('Cannot find cache by name %s.') % cachename
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        self.client.add_lun_to_cache(lun_id, cache_id)
        return cache_id

    def remove(self, cache_id, lun_id):
        self.client.remove_lun_from_cache(lun_id, cache_id)

    def update(self, cache_id, cachename, lun_id):
        cache_info = self.client.get_cache_info_by_id(cache_id)
        if cache_info['NAME'] == cachename:
            return

        self.remove(cache_id, lun_id)
        self.add(cachename, lun_id)

    def check_cache_valid(self, cachename):
        cache_id = self.client.get_cache_id_by_name(cachename)
        if not cache_id:
            msg = _("Cache %s doesn't exit.") % cachename
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heapq import heappop
from heapq import heappush
import time

from oslo_log import log as logging

from cinder.volume.drivers.dell_emc.powermax import utils


LOG = logging.getLogger(__name__)


class PowerMaxPerformance(object):
    """Performance Class for Dell EMC PowerMax volume drivers.

    It supports VMAX 3, All Flash and PowerMax arrays.
    """

    def __init__(self, rest, performance_config):
        self.rest = rest
        self.config = performance_config

    def set_performance_configuration(self, array_id, cinder_conf):
        """Set the performance configuration if details present in cinder.conf.

        :param array_id: the array serial number -- str
        :param cinder_conf: cinder configuration options -- dict
        """
        # Get performance registration, real-time registration, and collection
        # interval information for PowerMax array
        p_reg, rt_reg, c_int = self.get_array_registration_details(array_id)

        # Get load balance settings from cinder backend configuration
        lb_enabled = cinder_conf.safe_get(utils.LOAD_BALANCE)
        rt_enabled = cinder_conf.safe_get(utils.LOAD_BALANCE_RT)

        # Real-time
        if rt_enabled and not rt_reg:
            LOG.warning(
                "Real-time load balancing is enabled but array %(arr)s is not "
                "registered for real-time performance metrics collection. "
                "Diagnostic performance metrics will be used instead.",
                {'arr': array_id})
            rt_enabled = False

        # Load balancing enabled but array not registered for perf metrics
        if (lb_enabled or rt_enabled) and not p_reg:
            LOG.warning(
                "Load balancing is enabled but array %(arr)s is not "
                "registered for performance metrics collection. Reverting to "
                "default random Port and Port Group selection",
                {'arr': array_id})
            return {'load_balance': False}

        data_format = cinder_conf.safe_get(utils.PERF_DATA_FORMAT)
        if data_format.lower() not in ['average', 'avg', 'maximum', 'max']:
            LOG.warning("Incorrect data format '%(df)s', reverting to "
                        "default value 'Average'.", {'df': data_format})
            data_format = 'Average'

        if data_format.lower() in ['average', 'avg']:
            data_format = 'Average'
        elif data_format.lower() in ['maximum', 'max']:
            data_format = 'Maximum'

        # Get diagnostic metrics look back window
        lb_diagnostic = cinder_conf.safe_get(utils.LOAD_LOOKBACK)
        if not lb_diagnostic:
            LOG.warning(
                "Diagnostic look back window not set in cinder.conf, "
                "reverting to default value of 60 for most recent hour of "
                "metrics.")
            lb_diagnostic = 60
        elif lb_diagnostic < 0 or lb_diagnostic > 1440:
            LOG.warning(
                "Diagnostic look back window '%(lb)s' is not within the "
                "minimum and maximum range 0-1440, reverting to default "
                "value of 60 for most recent hour of metrics.", {
                    'lb': lb_diagnostic})
            lb_diagnostic = 60

        # Get real-time metrics look back window
        lb_real_time = cinder_conf.safe_get(utils.LOAD_LOOKBACK_RT)
        if rt_enabled:
            if not lb_real_time:
                LOG.warning(
                    "Real-time look back window not set in cinder.conf, "
                    "reverting to default value of 1 for for most recent "
                    "minute of metrics.")
                lb_real_time = 1
            elif lb_real_time < 1 or lb_real_time > 60:
                LOG.warning(
                    "Real-time look back window '%(lb)s' is not within the "
                    "minimum and maximum range 1-60, reverting to default "
                    "value of 1 for for most recent minute of metrics.", {
                        'lb': lb_real_time})
                lb_real_time = 1

        # Get Port Group metric for load calculation
        pg_metric = cinder_conf.safe_get(utils.PORT_GROUP_LOAD_METRIC)
        if not pg_metric:
            LOG.warning(
                "Port Group performance metric not set in cinder.conf, "
                "reverting to default metric 'PercentBusy'.")
            pg_metric = 'PercentBusy'
        elif pg_metric not in utils.PG_METRICS:
            LOG.warning(
                "Port Group performance metric selected for load "
                "balancing '%(pg_met)s' is not valid, reverting to "
                "default metric 'PercentBusy'.", {
                    'pg_met': pg_metric})
            pg_metric = 'PercentBusy'

        # Get Port metric for load calculation
        port_metric = cinder_conf.safe_get(utils.PORT_LOAD_METRIC)
        valid_port_metrics = (
            utils.PORT_RT_METRICS if rt_enabled else utils.PORT_METRICS)
        if not port_metric:
            LOG.warning(
                "Port performance metric not set in cinder.conf, "
                "reverting to default metric 'PercentBusy'.")
            port_metric = 'PercentBusy'
        elif port_metric not in valid_port_metrics:
            LOG.warning(
                "Port performance metric selected for load balancing "
                "'%(port_met)s' is not valid, reverting to default metric "
                "'PercentBusy'.", {'port_met': port_metric})
            port_metric = 'PercentBusy'

        self.config = {
            'load_balance': lb_enabled, 'load_balance_rt': rt_enabled,
            'perf_registered': p_reg, 'rt_registered': rt_reg,
            'collection_interval': c_int, 'data_format': data_format,
            'look_back': lb_diagnostic, 'look_back_rt': lb_real_time,
            'port_group_metric': pg_metric, 'port_metric': port_metric}

    def get_array_registration_details(self, array_id):
        """Get array performance registration details.

        :param array_id: the array serial number -- str
        :returns: performance registered, real-time registered,
                  collection interval -- bool, bool, int
        """
        LOG.info("Retrieving array %(arr)s performance registration details.",
                 {'arr': array_id})

        array_reg_uri = self.rest.build_uri(
            category=utils.PERFORMANCE, resource_level=utils.ARRAY_PERF,
            resource_type=utils.REG_DETAILS, resource_type_id=array_id,
            no_version=True)
        reg_details = self.rest.get_request(
            target_uri=array_reg_uri,
            resource_type='Array registration details')

        array_reg_info = reg_details.get(utils.REG_DETAILS_INFO)[0]
        perf_registered = array_reg_info.get(utils.DIAGNOSTIC)
        real_time_registered = array_reg_info.get(utils.REAL_TIME)
        collection_interval = array_reg_info.get(utils.COLLECTION_INT)

        return perf_registered, real_time_registered, collection_interval

    def get_array_performance_keys(self, array_id):
        """Get array performance keys (first and last available timestamps).

        :param array_id: the array serial number
        :returns: first date, last date -- int, int
        """
        LOG.debug("Retrieving array %(arr)s performance keys.",
                  {'arr': array_id})

        array_keys_uri = self.rest.build_uri(
            category=utils.PERFORMANCE, resource_level=utils.ARRAY_PERF,
            resource_type=utils.KEYS, no_version=True)

        array_keys = self.rest.get_request(
            target_uri=array_keys_uri, resource_type='Array performance keys')

        env_symm_info = array_keys.get(utils.ARRAY_INFO)
        f_date, l_date = None, None
        for symm in env_symm_info:
            if symm.get(utils.SYMM_ID) == array_id:
                f_date, l_date = symm.get(utils.F_DATE), symm.get(utils.L_DATE)

        return f_date, l_date

    @staticmethod
    def _get_look_back_window_interval_timestamp(l_date, lb_window):
        """Get first date value when calculated from last date and window.

        :param l_date: the last (most recent) timestamp -- int
        :param lb_window: the look back window in minutes -- int
        :returns: the first timestamp -- int
        """
        return l_date - (utils.ONE_MINUTE * lb_window)

    @staticmethod
    def _process_load(performance_data, metric):
        """Process the load for a given performance response, return average.

        :param performance_data: raw performance data from REST API -- dict
        :param metric: performance metric in use -- str
        :returns: range average, range total, interval count -- float, int, int
        """
        data = performance_data.get(utils.RESULT_LIST)
        result = data.get(utils.RESULT)

        total = 0
        for timestamp in result:
            total += timestamp.get(metric)

        return total / len(result), total, len(result)

    def _get_port_group_performance_stats(
            self, array_id, port_group_id, f_date, l_date, metric,
            data_format):
        """Get performance data for a given port group and performance metric.

        :param array_id: the array serial number -- str
        :param port_group_id: the port group id -- str
        :param f_date: first date for stats -- int
        :param l_date: last date for stats -- int
        :param metric: performance metric -- str
        :param data_format: performance data format -- str
        :returns: range average, range total, interval count -- float, float,
                                                                int
        """
        request_body = {
            utils.SYMM_ID: array_id, utils.PORT_GROUP_ID: port_group_id,
            utils.S_DATE: f_date, utils.E_DATE: l_date,
            utils.DATA_FORMAT: data_format, utils.METRICS: [metric]}

        port_group_uri = self.rest.build_uri(
            category=utils.PERFORMANCE, resource_level=utils.PORT_GROUP,
            resource_type=utils.METRICS, no_version=True)

        result = self.rest.post_request(
            port_group_uri, 'Port Group performance metrics',
            request_body)

        return self._process_load(result, metric)

    def _get_port_performance_stats(
            self, array_id, director_id, port_id, f_date, l_date, metric,
            data_format=None, real_time=False):
        """Get performance data for a given port and performance metric.

        :param array_id: the array serial number -- str
        :param director_id: the director id -- str
        :param port_id: the port id -- str
        :param f_date: first date for stats -- int
        :param l_date: last date for stats -- int
        :param metric: performance metric -- str
        :param data_format: performance data format -- str
        :param real_time: if metrics are real-time -- bool
        :returns: range average, range total, interval count -- float, float,
                                                                int
        """
        if real_time:
            target_uri = self.rest.build_uri(
                category=utils.PERFORMANCE, resource_level=utils.REAL_TIME,
                resource_type=utils.METRICS, no_version=True)
            res_type = 'real-time'
            dir_port = ('%(dir)s:%(port)s' % {'dir': director_id,
                                              'port': port_id})
            request_body = {
                utils.SYMM_ID: array_id, utils.INST_ID: dir_port,
                utils.S_DATE: f_date, utils.E_DATE: l_date,
                utils.CAT: utils.FE_PORT_RT, utils.METRICS: [metric]}

        else:
            target_uri = self.rest.build_uri(
                category=utils.PERFORMANCE, resource_level=utils.FE_PORT_DIAG,
                resource_type=utils.METRICS, no_version=True)
            res_type = 'diagnostic'
            request_body = {
                utils.SYMM_ID: array_id,
                utils.DIR_ID: director_id, utils.PORT_ID: port_id,
                utils.S_DATE: f_date, utils.E_DATE: l_date,
                utils.DATA_FORMAT: data_format, utils.METRICS: [metric]}

        resource = '%(res)s Port performance metrics' % {'res': res_type}
        result = self.rest.post_request(
            target_uri, resource, request_body)

        return self._process_load(result, metric)

    def process_port_group_load(
            self, array_id, port_groups, max_load=False):
        """Calculate the load for one or more port groups.

        :param array_id: the array serial number -- str
        :param port_groups: port group names -- list
        :param max_load: if max load port group should be returned -- bool
        :returns: low/max avg, metric, port group -- tuple(float, str, str)
        """
        LOG.info("Calculating array %(arr)s load for Port Groups %(pg)s.",
                 {'arr': array_id, 'pg': port_groups})

        data_format = self.config.get('data_format')
        lb_window = self.config.get('look_back')
        pg_metric = self.config.get('port_group_metric')

        __, l_date = self.get_array_performance_keys(array_id)
        f_date = self._get_look_back_window_interval_timestamp(
            l_date, lb_window)

        heap_low, heap_high = [], []

        start_time = time.time()
        for pg in port_groups:
            avg, total, cnt = self._get_port_group_performance_stats(
                array_id, pg, f_date, l_date, pg_metric, data_format)
            LOG.debug(
                "Port Group '%(pg)s' %(df)s %(met)s load for %(interval)s min "
                "interval: %(avg)s",
                {'pg': pg, 'df': data_format, 'met': pg_metric,
                 'interval': lb_window, 'avg': avg})

            # Add PG average to lowest load heap
            heappush(heap_low, (avg, pg_metric, pg))
            # Add inverse PG average to highest load heap
            heappush(heap_high, (-avg, pg_metric, pg))

        LOG.debug("Time taken to analyse Port Group performance: %(t)ss",
                  {'t': time.time() - start_time})

        return heappop(heap_high) if max_load else heappop(heap_low)

    def process_port_load(self, array_id, ports, max_load=False):
        """Calculate the load for one or more ports.

        :param array_id: the array serial number -- str
        :param ports: physical dir:port names -- list
        :param max_load: if max load port should be returned -- bool
        :returns: low/max avg, metric, port -- tuple(float, str, str)
        """
        LOG.info("Calculating array %(arr)s load for Ports %(port)s.",
                 {'arr': array_id, 'port': ports})

        rt_enabled = self.config.get('load_balance_rt')
        rt_registered = self.config.get('rt_registered')

        if rt_enabled and rt_registered:
            real_time, data_format = True, None
            lb_window = self.config.get('look_back_rt')
        else:
            real_time, data_format = False, self.config.get('data_format')
            lb_window = self.config.get('look_back')

        port_metric = self.config.get('port_metric')
        __, l_date = self.get_array_performance_keys(array_id)
        f_date = self._get_look_back_window_interval_timestamp(
            l_date, lb_window)

        heap_low, heap_high = [], []
        start_time = time.time()
        for port in ports:
            dir_id = port.split(':')[0]
            port_no = port.split(':')[1]

            avg, total, cnt = self._get_port_performance_stats(
                array_id, dir_id, port_no, f_date, l_date, port_metric,
                data_format, real_time=real_time)
            LOG.debug(
                "Port '%(dir)s:%(port)s' %(df)s %(met)s load for %(int)s min "
                "interval: %(avg)s",
                {'dir': dir_id, 'port': port_no,
                 'df': data_format if data_format else '',
                 'met': port_metric, 'int': lb_window, 'avg': avg})

            # Add PG average to lowest load heap
            heappush(heap_low, (avg, port_metric, port))
            # Add inverse PG average to highest load heap
            heappush(heap_high, (-avg, port_metric, port))

        LOG.debug("Time taken to analyse Port Group performance: %(t)ss",
                  {'t': time.time() - start_time})

        return heappop(heap_high) if max_load else heappop(heap_low)



# Copyright (c) 2011 Zadara Storage Inc.
# Copyright (c) 2011 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volume types manage extension."""

from oslo_utils import strutils
import six
from six.moves import http_client
import webob

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_types as volume_types_schema
from cinder.api import validation
from cinder.api.views import types as views_types
from cinder import exception
from cinder.i18n import _
from cinder.policies import volume_type as policy
from cinder import rpc
from cinder import utils
from cinder.volume import volume_types


class VolumeTypesManageController(wsgi.Controller):
    """The volume types API controller for the OpenStack API."""

    _view_builder_class = views_types.ViewBuilder

    @utils.if_notifications_enabled
    def _notify_volume_type_error(self, context, method, err,
                                  volume_type=None, id=None, name=None):
        payload = dict(
            volume_types=volume_type, name=name, id=id, error_message=err)
        rpc.get_notifier('volumeType').error(context, method, payload)

    @utils.if_notifications_enabled
    def _notify_volume_type_info(self, context, method, volume_type):
        payload = dict(volume_types=volume_type)
        rpc.get_notifier('volumeType').info(context, method, payload)

    @wsgi.action("create")
    @validation.schema(volume_types_schema.create)
    def _create(self, req, body):
        """Creates a new volume type."""
        context = req.environ['cinder.context']
        context.authorize(policy.MANAGE_POLICY)
        vol_type = body['volume_type']
        name = vol_type['name']
        description = vol_type.get('description')
        specs = vol_type.get('extra_specs', {})
        is_public = vol_type.get('os-volume-type-access:is_public', True)
        is_public = strutils.bool_from_string(is_public, strict=True)
        try:
            volume_types.create(context,
                                name,
                                specs,
                                is_public,
                                description=description)
            vol_type = volume_types.get_volume_type_by_name(context, name)
            req.cache_resource(vol_type, name='types')
            self._notify_volume_type_info(
                context, 'volume_type.create', vol_type)

        except exception.VolumeTypeExists as err:
            self._notify_volume_type_error(
                context, 'volume_type.create', err, volume_type=vol_type)
            raise webob.exc.HTTPConflict(explanation=six.text_type(err))
        except exception.VolumeTypeNotFoundByName as err:
            self._notify_volume_type_error(
                context, 'volume_type.create', err, name=name)
            # Not found exception will be handled at the wsgi level
            raise

        return self._view_builder.show(req, vol_type)

    @wsgi.action("update")
    @validation.schema(volume_types_schema.update)
    def _update(self, req, id, body):
        # Update description for a given volume type.
        context = req.environ['cinder.context']
        context.authorize(policy.MANAGE_POLICY)
        vol_type = body['volume_type']
        description = vol_type.get('description')
        name = vol_type.get('name')
        is_public = vol_type.get('is_public')

        if is_public is not None:
            is_public = strutils.bool_from_string(is_public, strict=True)

        # If name specified, name can not be empty.
        if name and len(name.strip()) == 0:
            msg = _("Volume type name can not be empty.")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        # Name, description and is_public can not be None.
        # Specify one of them, or a combination thereof.
        if name is None and description is None and is_public is None:
            msg = _("Specify volume type name, description, is_public or "
                    "a combination thereof.")
            raise webob.exc.HTTPBadRequest(explanation=msg)

        try:
            volume_types.update(context, id, name, description,
                                is_public=is_public)
            # Get the updated
            vol_type = volume_types.get_volume_type(context, id)
            req.cache_resource(vol_type, name='types')
            self._notify_volume_type_info(
                context, 'volume_type.update', vol_type)

        except exception.VolumeTypeNotFound as err:
            self._notify_volume_type_error(
                context, 'volume_type.update', err, id=id)
            # Not found exception will be handled at the wsgi level
            raise
        except exception.VolumeTypeExists as err:
            self._notify_volume_type_error(
                context, 'volume_type.update', err, volume_type=vol_type)
            raise webob.exc.HTTPConflict(explanation=six.text_type(err))
        except exception.VolumeTypeUpdateFailed as err:
            self._notify_volume_type_error(
                context, 'volume_type.update', err, volume_type=vol_type)
            raise webob.exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        return self._view_builder.show(req, vol_type)

    @wsgi.action("delete")
    def _delete(self, req, id):
        """Deletes an existing volume type."""
        context = req.environ['cinder.context']
        context.authorize(policy.MANAGE_POLICY)

        try:
            vol_type = volume_types.get_volume_type(context, id)
            volume_types.destroy(context, vol_type['id'])
            self._notify_volume_type_info(
                context, 'volume_type.delete', vol_type)
        except exception.VolumeTypeInUse as err:
            self._notify_volume_type_error(
                context, 'volume_type.delete', err, volume_type=vol_type)
            msg = _('Target volume type is still in use.')
            raise webob.exc.HTTPBadRequest(explanation=msg)
        except exception.VolumeTypeNotFound as err:
            self._notify_volume_type_error(
                context, 'volume_type.delete', err, id=id)
            # Not found exception will be handled at the wsgi level
            raise
        except (exception.VolumeTypeDeletionError,
                exception.VolumeTypeDefaultDeletionError) as err:
            self._notify_volume_type_error(
                context, 'volume_type.delete', err, volume_type=vol_type)
            raise webob.exc.HTTPBadRequest(explanation=err.msg)
        except exception.VolumeTypeDefaultMisconfiguredError as err:
            self._notify_volume_type_error(
                context, 'volume_type.delete', err, volume_type=vol_type)
            raise webob.exc.HTTPInternalServerError(explanation=err.msg)

        return webob.Response(status_int=http_client.ACCEPTED)


class Types_manage(extensions.ExtensionDescriptor):
    """Types manage support."""

    name = "TypesManage"
    alias = "os-types-manage"
    updated = "2011-08-24T00:00:00+00:00"

    def get_controller_extensions(self):
        controller = VolumeTypesManageController()
        extension = extensions.ControllerExtension(self, 'types', controller)
        return [extension]



# Copyright 2016 Dell Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Consistency group volume driver interface.
"""

from cinder.interface import base


class VolumeConsistencyGroupDriver(base.CinderInterface):
    """Interface for drivers that support consistency groups."""

    def create_consistencygroup(self, context, group):
        """Creates a consistencygroup.

        :param context: the context of the caller.
        :param group: the dictionary of the consistency group to be created.
        :returns: model_update

        model_update will be in this format: {'status': xxx, ......}.

        If the status in model_update is 'error', the manager will throw
        an exception and it will be caught in the try-except block in the
        manager. If the driver throws an exception, the manager will also
        catch it in the try-except block. The group status in the db will
        be changed to 'error'.

        For a successful operation, the driver can either build the
        model_update and return it or return None. The group status will
        be set to 'available'.
        """

    def create_consistencygroup_from_src(self, context, group, volumes,
                                         cgsnapshot=None, snapshots=None,
                                         source_cg=None, source_vols=None):
        """Creates a consistencygroup from source.

        :param context: the context of the caller.
        :param group: the dictionary of the consistency group to be created.
        :param volumes: a list of volume dictionaries in the group.
        :param cgsnapshot: the dictionary of the cgsnapshot as source.
        :param snapshots: a list of snapshot dictionaries in the cgsnapshot.
        :param source_cg: the dictionary of a consistency group as source.
        :param source_vols: a list of volume dictionaries in the source_cg.
        :returns: model_update, volumes_model_update

        The source can be cgsnapshot or a source cg.

        param volumes is retrieved directly from the db. It is a list of
        cinder.db.sqlalchemy.models.Volume to be precise. It cannot be
        assigned to volumes_model_update. volumes_model_update is a list of
        dictionaries. It has to be built by the driver. An entry will be
        in this format: {'id': xxx, 'status': xxx, ......}. model_update
        will be in this format: {'status': xxx, ......}.

        To be consistent with other volume operations, the manager will
        assume the operation is successful if no exception is thrown by
        the driver. For a successful operation, the driver can either build
        the model_update and volumes_model_update and return them or
        return None, None.
        """

    def delete_consistencygroup(self, context, group, volumes):
        """Deletes a consistency group.

        :param context: the context of the caller.
        :param group: the dictionary of the consistency group to be deleted.
        :param volumes: a list of volume dictionaries in the group.
        :returns: model_update, volumes_model_update

        param volumes is retrieved directly from the db. It is a list of
        cinder.db.sqlalchemy.models.Volume to be precise. It cannot be
        assigned to volumes_model_update. volumes_model_update is a list of
        dictionaries. It has to be built by the driver. An entry will be
        in this format: {'id': xxx, 'status': xxx, ......}. model_update
        will be in this format: {'status': xxx, ......}.

        The driver should populate volumes_model_update and model_update
        and return them.

        The manager will check volumes_model_update and update db accordingly
        for each volume. If the driver successfully deleted some volumes
        but failed to delete others, it should set statuses of the volumes
        accordingly so that the manager can update db correctly.

        If the status in any entry of volumes_model_update is 'error_deleting'
        or 'error', the status in model_update will be set to the same if it
        is not already 'error_deleting' or 'error'.

        If the status in model_update is 'error_deleting' or 'error', the
        manager will raise an exception and the status of the group will be
        set to 'error' in the db. If volumes_model_update is not returned by
        the driver, the manager will set the status of every volume in the
        group to 'error' in the except block.

        If the driver raises an exception during the operation, it will be
        caught by the try-except block in the manager. The statuses of the
        group and all volumes in it will be set to 'error'.

        For a successful operation, the driver can either build the
        model_update and volumes_model_update and return them or
        return None, None. The statuses of the group and all volumes
        will be set to 'deleted' after the manager deletes them from db.
        """

    def update_consistencygroup(self, context, group,
                                add_volumes=None, remove_volumes=None):
        """Updates a consistency group.

        :param context: the context of the caller.
        :param group: the dictionary of the consistency group to be updated.
        :param add_volumes: a list of volume dictionaries to be added.
        :param remove_volumes: a list of volume dictionaries to be removed.
        :returns: model_update, add_volumes_update, remove_volumes_update

        model_update is a dictionary that the driver wants the manager
        to update upon a successful return. If None is returned, the manager
        will set the status to 'available'.

        add_volumes_update and remove_volumes_update are lists of dictionaries
        that the driver wants the manager to update upon a successful return.
        Note that each entry requires a {'id': xxx} so that the correct
        volume entry can be updated. If None is returned, the volume will
        remain its original status. Also note that you cannot directly
        assign add_volumes to add_volumes_update as add_volumes is a list of
        cinder.db.sqlalchemy.models.Volume objects and cannot be used for
        db update directly. Same with remove_volumes.

        If the driver throws an exception, the status of the group as well as
        those of the volumes to be added/removed will be set to 'error'.
        """

    def create_cgsnapshot(self, context, cgsnapshot, snapshots):
        """Creates a cgsnapshot.

        :param context: the context of the caller.
        :param cgsnapshot: the dictionary of the cgsnapshot to be created.
        :param snapshots: a list of snapshot dictionaries in the cgsnapshot.
        :returns: model_update, snapshots_model_update

        param snapshots is retrieved directly from the db. It is a list of
        cinder.db.sqlalchemy.models.Snapshot to be precise. It cannot be
        assigned to snapshots_model_update. snapshots_model_update is a list
        of dictionaries. It has to be built by the driver. An entry will be
        in this format: {'id': xxx, 'status': xxx, ......}. model_update
        will be in this format: {'status': xxx, ......}.

        The driver should populate snapshots_model_update and model_update
        and return them.

        The manager will check snapshots_model_update and update db accordingly
        for each snapshot. If the driver successfully deleted some snapshots
        but failed to delete others, it should set statuses of the snapshots
        accordingly so that the manager can update db correctly.

        If the status in any entry of snapshots_model_update is 'error', the
        status in model_update will be set to the same if it is not already
        'error'.

        If the status in model_update is 'error', the manager will raise an
        exception and the status of cgsnapshot will be set to 'error' in the
        db. If snapshots_model_update is not returned by the driver, the
        manager will set the status of every snapshot to 'error' in the except
        block.

        If the driver raises an exception during the operation, it will be
        caught by the try-except block in the manager and the statuses of
        cgsnapshot and all snapshots will be set to 'error'.

        For a successful operation, the driver can either build the
        model_update and snapshots_model_update and return them or
        return None, None. The statuses of cgsnapshot and all snapshots
        will be set to 'available' at the end of the manager function.
        """

    def delete_cgsnapshot(self, context, cgsnapshot, snapshots):
        """Deletes a cgsnapshot.

        :param context: the context of the caller.
        :param cgsnapshot: the dictionary of the cgsnapshot to be deleted.
        :param snapshots: a list of snapshot dictionaries in the cgsnapshot.
        :returns: model_update, snapshots_model_update

        param snapshots is retrieved directly from the db. It is a list of
        cinder.db.sqlalchemy.models.Snapshot to be precise. It cannot be
        assigned to snapshots_model_update. snapshots_model_update is a list
        of dictionaries. It has to be built by the driver. An entry will be
        in this format: {'id': xxx, 'status': xxx, ......}. model_update
        will be in this format: {'status': xxx, ......}.

        The driver should populate snapshots_model_update and model_update
        and return them.

        The manager will check snapshots_model_update and update db accordingly
        for each snapshot. If the driver successfully deleted some snapshots
        but failed to delete others, it should set statuses of the snapshots
        accordingly so that the manager can update db correctly.

        If the status in any entry of snapshots_model_update is
        'error_deleting' or 'error', the status in model_update will be set to
        the same if it is not already 'error_deleting' or 'error'.

        If the status in model_update is 'error_deleting' or 'error', the
        manager will raise an exception and the status of cgsnapshot will be
        set to 'error' in the db. If snapshots_model_update is not returned by
        the driver, the manager will set the status of every snapshot to
        'error' in the except block.

        If the driver raises an exception during the operation, it will be
        caught by the try-except block in the manager and the statuses of
        cgsnapshot and all snapshots will be set to 'error'.

        For a successful operation, the driver can either build the
        model_update and snapshots_model_update and return them or
        return None, None. The statuses of cgsnapshot and all snapshots
        will be set to 'deleted' after the manager deletes them from db.
        """



# Copyright (c) 2015 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Exception definitions.
"""

from oslo_vmware import exceptions

from cinder.i18n import _


class InvalidAdapterTypeException(exceptions.VMwareDriverException):
    """Thrown when the disk adapter type is invalid."""
    msg_fmt = _("Invalid disk adapter type: %(invalid_type)s.")


class InvalidDiskTypeException(exceptions.VMwareDriverException):
    """Thrown when the disk type is invalid."""
    msg_fmt = _("Invalid disk type: %(disk_type)s.")


class VirtualDiskNotFoundException(exceptions.VMwareDriverException):
    """Thrown when virtual disk is not found."""
    msg_fmt = _("There is no virtual disk device.")


class ProfileNotFoundException(exceptions.VMwareDriverException):
    """Thrown when the given storage profile cannot be found."""
    msg_fmt = _("Storage profile: %(storage_profile)s not found.")


class NoValidDatastoreException(exceptions.VMwareDriverException):
    """Thrown when there are no valid datastores."""
    msg_fmt = _("There are no valid datastores.")


class ClusterNotFoundException(exceptions.VMwareDriverException):
    """Thrown when the given cluster cannot be found."""
    msg_fmt = _("Compute cluster: %(cluster)s not found.")


class NoValidHostException(exceptions.VMwareDriverException):
    """Thrown when there are no valid ESX hosts."""
    msg_fmt = _("There are no valid ESX hosts.")


class TemplateNotFoundException(exceptions.VMwareDriverException):
    """Thrown when template cannot be found."""
    msg_fmt = _("Template cannot be found at path: %(path)s.")


class SnapshotNotFoundException(exceptions.VMwareDriverException):
    """Thrown when the backend snapshot cannot be found."""
    msg_fmt = _("Snapshot: %(name)s not found.")
